"""Module with a data class object which represents CRT information.

This copies the internal structure of :class:`larcv.CRTHit`.
"""

from dataclasses import dataclass

import numpy as np

from .base import PosDataBase

__all__ = ["CRTHit"]


@dataclass(eq=False)
class CRTHit(PosDataBase):
    """CRT hit information.

    Attributes
    ----------
    id : int
        Index of the CRT hit in the list
    plane : int
        Index of the CRT tagger that registered the hit
    tagger : str
        Name of the CRT tagger that registered the hit
    feb_id : np.ndarray
        Address of the FEB board stored as a list of bytes (uint8)
    ts0_s : int
        Absolute time from White Rabbit (seconds component)
    ts0_ns : float
        Absolute time from White Rabbit (nanoseconds component)
    ts0_s_corr : float
        Unclear in the documentation, placeholder at this point
    ts0_ns_corr : float
        Unclear in the documentation, placeholder at this point
    ts1_ns : float
        Time relative to the trigger (nanoseconds component)
    time : float
        Alias for ts1_ns, but scaled to microseconds
    total_pe : float
        Total number of PE in the CRT hit
    pe_per_ch : np.ndarray
        Number of PEs per FEB channel
    center : np.ndarray
        Barycenter of the CRT hit in detector coordinates
    width : np.ndarray
        Uncertainty on the barycenter of the CRT hit in detector coordinates
    units : str
        Units in which the position attributes are expressed
    """

    id: int = -1
    plane: int = -1
    tagger: str = ""
    feb_id: np.ndarray = None
    ts0_s: int = -1
    ts0_ns: float = -1.0
    ts0_s_corr: float = -1.0
    ts0_ns_corr: float = -1.0
    ts1_ns: float = -1.0
    time: float = -1.0
    total_pe: float = -1.0
    # pe_per_ch: np.ndarray = None
    center: np.ndarray = None
    width: np.ndarray = None
    units: str = "cm"

    # Fixed-length attributes
    _fixed_length_attrs = (("center", 3), ("width", 3))

    # Variable-length attributes
    _var_length_attrs = (("feb_id", np.ubyte),)

    # Attributes specifying coordinates
    _pos_attrs = ("position",)

    # Attributes specifying vector components
    _vec_attrs = ("width",)

    # String attributes
    _str_attrs = ("tagger", "units")

    # Index attributes
    _index_attrs = ("id",)

    @property
    def time(self):
        """Time w.r.t. to the trigger.

        Returns
        -------
        float
            Time of the CRT hit w.r.t. to the trigger in microseconds.
        """
        return self.ts1_ns / 1000.0

    @time.setter
    def time(self, time):
        pass

    @classmethod
    def from_larcv(cls, crthit):
        """Builds and returns a CRTHit object from a LArCV CRTHit object.

        Parameters
        ----------
        crthit : larcv.CRTHit
            LArCV-format CRT hit

        Returns
        -------
        CRTHit
            CRT hit object
        """
        # Get the physical center and width of the CRT hit
        axes = ("x", "y", "z")
        center = np.array([getattr(crthit, f"{a}_pos")() for a in axes])
        width = np.array([getattr(crthit, f"{a}_err")() for a in axes])

        # Convert the FEB address to a list of bytes
        feb_id = np.array([ord(c) for c in crthit.feb_id()], dtype=np.ubyte)

        # Get the number of PEs per FEB channel
        # TODO: This is a dictionary of dictionaries, need to figure out
        # how to unpack in a sensible manner

        return cls(
            id=crthit.id(),
            plane=crthit.plane(),
            tagger=crthit.tagger(),
            feb_id=feb_id,
            ts0_s=crthit.ts0_s(),
            ts0_ns=crthit.ts0_ns(),
            ts0_s_corr=crthit.ts0_s_corr(),
            ts0_ns_corr=crthit.ts0_ns_corr(),
            ts1_ns=crthit.ts1_ns(),
            total_pe=crthit.peshit(),
            center=center,
            width=width,
        )



"""Numba JIT compiled implementation of graph routines.

In particular, this module supports the CSR data structure and derived methods,
which tremendously speeds up graph construction and computation in Numba.
"""

import numba as nb
import numpy as np

from .distance import METRICS, cdist, chebyshev, cityblock, minkowski, sqeuclidean

CSR_DTYPE = (
    ("num_nodes", nb.int64),
    ("neighbors", nb.int64[:]),
    ("offsets", nb.int64[:]),
)


@nb.experimental.jitclass(CSR_DTYPE)
class CSRGraph:
    """Numba-enabled compressed Sparse Row (CSR) representation of a sparse matrix.

    Attributes
    ----------
    neighbors : np.ndarray
        (E) List of node neighbors in a compressed array
    offsets : np.ndarray
        (N+1) Per-node slicing boundaries to query each node neighborhood
    num_nodes : int
        Number of nodes in the graph, N
    """

    def __init__(
        self, neighbors: nb.int64[:], offsets: nb.int64[:], num_nodes: nb.int64
    ):
        """Construct the Compressed Sparse Row (CSR) representation of a
        sparse matrix based on a list of nodes and edges.

        Parameters
        ----------
        neighbors : np.ndarray
            (E) List of node neighbors in a compressed array
        offsets : np.ndarray
            (N+1) Per-node slicing boundaries to query each node neighborhood
        num_nodes : int
            Number of nodes in the graph, N
        """
        self.neighbors = neighbors
        self.offsets = offsets
        self.num_nodes = num_nodes

    def __getitem__(self, node_id: nb.int64):
        """Get the list of neighbors associated with a node.

        Parameters
        ----------
        node_id : int
            Node index i

        Returns
        -------
        np.ndarray
            List of neighbors associated with node i
        """
        start, end = self.offsets[node_id], self.offsets[node_id + 1]
        return self.neighbors[start:end]

    def num_neighbors(self, node_id: nb.int64):
        """Returns the number of neighbors of a node.

        Parameters
        ----------
        node_id : int
            Node index i

        Returns
        -------
        int
            Number of neighbors of node i
        """
        start, end = self.offsets[node_id], self.offsets[node_id + 1]
        return end - start


@nb.njit
def csr_graph(
    edge_index: nb.int64[:, :], num_nodes: nb.int64, directed: nb.boolean = True
) -> CSR_DTYPE:
    """Construct the Compressed Sparse Row (CSR) representation of a sparse
    matrix based on a list of nodes and edges.

    Parameters
    ----------
    edge_index : np.ndarray
        (E, 2) List of active edge indices in the graph
    num_nodes : int
        Number of nodes in the graph, N
    directed : bool
        Whether the input graph is directed or not
    """
    # Count the number of connections per node
    counts = np.zeros(num_nodes, dtype=np.int64)
    for s, t in edge_index:
        counts[s] += 1
        if not directed:
            counts[t] += 1

    # Build the offsets array
    offsets = np.empty(num_nodes + 1, dtype=np.int64)
    offsets[0] = 0
    for i in range(num_nodes):
        offsets[i + 1] = offsets[i] + counts[i]

    # Build the neighbors array
    neighbors = np.empty(offsets[-1], dtype=np.int64)
    fill = np.zeros(num_nodes, dtype=np.int64)
    for s, t in edge_index:
        idx = offsets[s] + fill[s]
        neighbors[idx] = t
        fill[s] += 1
        if not directed:
            idx = offsets[t] + fill[t]
            neighbors[idx] = s
            fill[t] += 1

    # Initialize the CSR graph
    return CSRGraph(neighbors, offsets, num_nodes)


@nb.njit(cache=True)
def connected_components(
    edge_index: nb.int64[:, :],
    num_nodes: nb.int64,
    min_samples: nb.int64 = 1,
    directed: nb.boolean = True,
) -> nb.int64[:]:
    """Find connected components.

    Parameters
    ----------
    edge_index : np.ndarray
        (E, 2) List of active edge indices in the graph
    num_nodes : int
        Number of nodes in the graph, N
    directed : bool, default True
        Whether the input graph is directed or not

    Returns
    -------
    np.ndarray
        (N) Cluster label associated with each node
    """
    # Initialize the CSR data structure
    graph = csr_graph(edge_index, num_nodes, directed)

    # Initialize output
    labels = np.arange(graph.num_nodes)
    visited = np.zeros(graph.num_nodes, dtype=nb.boolean)
    component = np.empty(graph.num_nodes, dtype=nb.int64)
    comp_idx = np.empty(1, dtype=nb.int64)  # Acts as pointer

    # Loop through all nodes and start DFS from unvisited nodes
    label = 0
    min_neighbors = min_samples - 1
    for node in range(graph.num_nodes):
        if not visited[node]:
            if graph.num_neighbors(node) > min_neighbors:
                # Perform DFS and collect all nodes in this connected component
                comp_idx[0] = 0
                dfs_iterative(graph, visited, node, component, comp_idx)

                # Collect all nodes that belong to the same connected component
                for i in range(comp_idx[0]):
                    labels[component[i]] = label

            else:
                # Relabel solitary nodes to maintain ordering
                labels[node] = label

            # Increment label
            label += 1

    return labels


@nb.njit(cache=True)
def dfs(
    graph: CSR_DTYPE,
    visited: nb.boolean[:],
    node: nb.int64,
    component: nb.int64[:],
    comp_idx: nb.int64[:],
):
    """Does a depth-first search and builds a connected component.

    Parameters
    ----------
    graph : CSRGraph
        CSR representation of a graph
    visitied : np.ndarray
        (N) Boolean array which specified weather a node has been visited or not.
    node : int
        Current node index
    component : np.ndarray
        (N) Current component (padded)
    comp_idx : np.ndarray
        Current component index (pointer)

    Notes
    -----
    This implementation is recursive, which is the fastest implementation but
    silently throws segementation faults if the maximum recursion depth is
    reached. The :func:`dfs_iterative` function is safer, but slightly slower.
    """
    # Mark the node as visited, incremant pointer
    visited[node] = True
    component[comp_idx[0]] = node
    comp_idx[0] += 1

    # Traverse all the neighbors of the node
    for neighbor in graph[node]:
        if not visited[neighbor]:
            dfs(graph, visited, neighbor, component, comp_idx)


@nb.njit(cache=True)
def dfs_iterative(
    graph: CSR_DTYPE,
    visited: nb.boolean[:],
    start_node: nb.int64,
    component: nb.int64[:],
    comp_idx: nb.int64[:],
):
    """Does a depth-first search and builds a connected component.

    Parameters
    ----------
    graph : CSRGraph
        CSR representation of a graph
    visitied : np.ndarray
        (N) Boolean array which specified weather a node has been visited or not.
    node : int
        Current node index
    component : np.ndarray
        (N) Current component (padded)
    comp_idx : np.ndarray
        Current component index (pointer)

    Notes
    -----
    This implementation is iterative and does not suffer from the recursion
    depth maximum issue which affects the recursive version, at a small cost
    to the overall execution speed.
    """
    # Initialize a node stack (fixed size)
    stack = np.empty(graph.num_nodes, dtype=nb.int64)
    stack[0] = start_node
    stack_idx = 1

    visited[start_node] = True

    # Loop until there is no more node to visit
    while stack_idx > 0:
        stack_idx -= 1
        node = stack[stack_idx]

        component[comp_idx[0]] = node
        comp_idx[0] += 1

        for neighbor in graph[node]:
            if not visited[neighbor]:
                visited[neighbor] = True
                stack[stack_idx] = neighbor
                stack_idx += 1


@nb.njit(cache=True)
def radius_graph(
    x: nb.float32[:, :],
    radius: nb.float32,
    metric_id: nb.int64 = METRICS["euclidean"],
    p: nb.float32 = 2.0,
) -> nb.int64[:, :]:
    """Builds an undirected radius graph.

    This function generates a list of edges in a graph which connects all nodes
    which live within some radius R of each other.

    Parameters
    ----------
    x : np.ndarray
        (N, 3) array of node coordinates
    radius : float
        Radius within which to build connections in the graph
    metric_id : int, default 2 (Euclidean)
        Distance metric enumerator
    p : float, default 2.
        p-norm factor for the Minkowski metric, if used

    Returns
    -------
    np.ndarray
        (E, 2) array of edges in the radius graph
    """
    # Determine the distance function to use. If the metric is Euclidean, it
    # is cheaper to square the radius and use the squared Euclidean metric
    if metric_id == np.int64(0):
        return _radius_graph_minkowski(x, radius, p)
    elif metric_id == np.int64(1):
        return _radius_graph_cityblock(x, radius)
    elif metric_id == np.int64(2):
        radius = radius * radius
        return _radius_graph_sqeuclidean(x, radius)
    elif metric_id == np.int64(3):
        return _radius_graph_sqeuclidean(x, radius)
    elif metric_id == np.int64(4):
        return _radius_graph_chebyshev(x, radius)
    else:
        raise ValueError("Distance metric not recognized.")


@nb.njit(cache=True)
def _radius_graph_minkowski(
    x: nb.float32[:, :], radius: nb.float32, p: nb.float32
) -> nb.float32[:, :]:
    # Initialize a data structure to hold edges
    num_nodes = len(x)
    max_edges = num_nodes * (num_nodes - 1) // 2
    edge_index = np.empty((max_edges, 2), dtype=np.int64)

    # Loop over pairs of nodes, ass edges if the distance fits the bill
    edge_count = 0
    for i in range(num_nodes):
        for j in range(i + 1, num_nodes):
            if minkowski(x[i], x[j], p) <= radius:
                edge_index[edge_count, 0], edge_index[edge_count, 1] = i, j
                edge_count += 1

    return edge_index[:edge_count]


@nb.njit(cache=True)
def _radius_graph_cityblock(
    x: nb.float32[:, :], radius: nb.float32
) -> nb.float32[:, :]:
    # Initialize a data structure to hold edges
    num_nodes = len(x)
    max_edges = num_nodes * (num_nodes - 1) // 2
    edge_index = np.empty((max_edges, 2), dtype=np.int64)

    # Loop over pairs of nodes, ass edges if the distance fits the bill
    edge_count = 0
    for i in range(num_nodes):
        for j in range(i + 1, num_nodes):
            if cityblock(x[i], x[j]) <= radius:
                edge_index[edge_count, 0], edge_index[edge_count, 1] = i, j
                edge_count += 1

    return edge_index[:edge_count]


@nb.njit(cache=True)
def _radius_graph_sqeuclidean(
    x: nb.float32[:, :], radius: nb.float32
) -> nb.float32[:, :]:
    # Initialize a data structure to hold edges
    num_nodes = len(x)
    max_edges = num_nodes * (num_nodes - 1) // 2
    edge_index = np.empty((max_edges, 2), dtype=np.int64)

    # Loop over pairs of nodes, ass edges if the distance fits the bill
    edge_count = 0
    for i in range(num_nodes):
        for j in range(i + 1, num_nodes):
            if sqeuclidean(x[i], x[j]) <= radius:
                edge_index[edge_count, 0], edge_index[edge_count, 1] = i, j
                edge_count += 1

    return edge_index[:edge_count]


@nb.njit(cache=True)
def _radius_graph_chebyshev(
    x: nb.float32[:, :], radius: nb.float32
) -> nb.float32[:, :]:
    # Initialize a data structure to hold edges
    num_nodes = len(x)
    max_edges = num_nodes * (num_nodes - 1) // 2
    edge_index = np.empty((max_edges, 2), dtype=np.int64)

    # Loop over pairs of nodes, ass edges if the distance fits the bill
    edge_count = 0
    for i in range(num_nodes):
        for j in range(i + 1, num_nodes):
            if chebyshev(x[i], x[j]) <= radius:
                edge_index[edge_count, 0], edge_index[edge_count, 1] = i, j
                edge_count += 1

    return edge_index[:edge_count]


@nb.njit(cache=True)
def union_find(
    edge_index: nb.int64[:, :], count: nb.int64, return_inverse: bool = True
) -> nb.int64[:]:
    """Numba implementation of the Union-Find algorithm.

    This function assigns a group to each node in a graph, provided
    a set of edges connecting the nodes together.

    Parameters
    ----------
    edge_index : np.ndarray
        (E, 2) List of edges (sparse adjacency matrix)
    count : int
        Number of nodes in the graph, C
    return_inverse : bool, default True
        Make sure the group IDs range from 0 to N_groups-1

    Returns
    -------
    np.ndarray
        (C) Group assignments for each of the nodes in the graph
    Dict[int, np.ndarray]
        Dictionary which maps groups to indexes
    """
    labels = np.arange(count)
    groups = {i: np.array([i]) for i in labels}
    for e in edge_index:
        li, lj = labels[e[0]], labels[e[1]]
        if li != lj:
            labels[groups[lj]] = li
            groups[li] = np.concatenate((groups[li], groups[lj]))
            del groups[lj]

    if return_inverse:
        mask = np.zeros(count, dtype=np.bool_)
        mask[labels] = True
        mapping = np.empty(count, dtype=labels.dtype)
        mapping[mask] = np.arange(np.sum(mask))
        labels = mapping[labels]

    return labels, groups



"""Module with a data class object which represents rasterized images metadata.

This copies the internal structure of either :class:`larcv.ImageMeta` for 2D
images or :class:`larcv.Voxel3DMeta` for 3D images.
"""

from dataclasses import dataclass

import numpy as np

from .base import DataBase

__all__ = ["Meta"]


@dataclass(eq=False)
class Meta(DataBase):
    """Meta information about a rasterized image.

    Attributes
    ----------
    lower : np.ndarray
        (2/3) Array of image lower bounds in detector coordinates (cm)
    upper : np.ndarray
        (2/3) Array of image upper bounds in detector coordinates (cm)
    size : np.ndarray
        (2/3) Array of pixel size in each dimension (cm)
    count : np.ndarray
        (2/3) Array of pixel count in each dimension
    """

    lower: np.ndarray = None
    upper: np.ndarray = None
    size: np.ndarray = None
    count: np.ndarray = None

    # Fixed-length attributes
    _fixed_length_attrs = (
        ("lower", 3),
        ("upper", 3),
        ("size", 3),
        ("count", (3, np.int64)),
    )

    # Attributes specifying vector components
    _vec_attrs = ("lower", "upper", "size", "count")

    @property
    def dimension(self):
        """Number of dimensions in the image.

        Returns
        -------
        int
            Number of dimensions in the image
        """
        return len(self.lower)

    @property
    def num_elements(self):
        """Total number of pixel in the image.

        Returns
        -------
        int
            Total number of pixel in the image.
        """
        return int(np.prod(self.count))

    def index(self, coords):
        """Unique pix associated with individual axis indexes.

        Parameters
        ----------
        coords : np.ndarray
            (N, 2/3) Input pixel indices

        Returns
        -------
        np.ndarray
            (N) Unique pixel index per input pixel
        """
        mult = np.empty(self.dimension, dtype=self.count.dtype)
        for i in range(self.dimension):
            mult[i] = np.prod(self.count[i + 1 :])

        return np.dot(coords, mult).astype(np.int64)

    def to_cm(self, coords, center=False):
        """Converts pixel coordinates to detector coordinates in cm.

        Parameters
        ----------
        coords : np.ndarray
            (N, 2/3) Input pixel coordinates
        center : bool, default False
            If `True`, offset the input coordinates by half a pixel size. This
            makes sense to provide unbiased coordinates when converting indexes.

        Returns
        -------
        np.ndarray
            Detector coordinates in cm
        """
        return self.lower + (coords + 0.5 * center) * self.size

    def to_px(self, coords, floor=False):
        """Converts detector coordinates in cm to pixel coordinates.

        Parameters
        ----------
        coords : np.ndarray
            (N, 2/3) Input detector coordinates
        floor : bool, default False
            If `True`, converts pixel coordinates to indexes (floor function)

        Returns
        -------
        np.ndarray
            Pixel coordinates
        """
        if floor:
            return np.floor((coords - self.lower) / self.size)

        return (coords - self.lower) / self.size

    @classmethod
    def from_larcv(cls, meta):
        """Builds and returns a Meta object from a LArCV 2D metadata object.

        Parameters
        ----------
        meta : Union[larcv.ImageMeta, larcv.Voxel3DMeta]
            LArCV-format 2D metadata

        Returns
        -------
        Meta
            Metadata object
        """
        if hasattr(meta, "pos_z"):
            lower = np.array(
                [meta.min_x(), meta.min_y(), meta.min_z()], dtype=np.float32
            )
            upper = np.array(
                [meta.max_x(), meta.max_y(), meta.max_z()], dtype=np.float32
            )
            size = np.array(
                [meta.size_voxel_x(), meta.size_voxel_y(), meta.size_voxel_z()],
                dtype=np.float32,
            )
            count = np.array(
                [meta.num_voxel_x(), meta.num_voxel_y(), meta.num_voxel_z()],
                dtype=np.int64,
            )

        else:
            lower = np.array([meta.min_x(), meta.min_y()], dtype=np.float32)
            upper = np.array([meta.max_x(), meta.max_y()], dtype=np.float32)
            size = np.array([meta.pixel_height(), meta.pixel_width()], dtype=np.float32)
            count = np.array([meta.rows(), meta.cols()], dtype=np.int64)

        return cls(lower=lower, upper=upper, size=size, count=count)



"""Module which contains enumerated variables shared across the project."""

from enum import IntEnum

from .globals import *

__all__ = ["enum_factory"]


def enum_factory(enum, value):
    """Parses an enumerated object from string name(s) to value(s).

    Parameters
    ----------
    enum : str
        Name of the enumerated type
    value : Union[str, List[str]]
        Name or names of the enumerated objects (from config)

    Returns
    -------
    Union[int, List[int]]
        Value or values of the enumerated objects
    """
    # Get the enumerated type
    ENUM_DICT = {"cluster": ClusterLabelEnum, "shape": ShapeEnum, "pid": PIDEnum}
    assert enum in ENUM_DICT, (
        "Enumerated type not recognized: {enum}. Must be one of "
        "{list(ENUM_DICT.keys())}."
    )
    enum = ENUM_DICT[enum]

    # Translate enumerated strings into values
    if isinstance(value, str):
        if not hasattr(enum, value.upper()):
            raise ValueError(
                f"Enumerated object not recognized: {value}. Must be one "
                f"of {[e.name for e in enum]}."
            )

        return getattr(enum, value.upper()).value

    else:
        values = []
        for v in value:
            if not hasattr(enum, v.upper()):
                raise ValueError(
                    f"Enumerated object not recognized: {v}. Must be one "
                    f"of {[e.name for e in enum]}."
                )
            values.append(getattr(enum, v.upper()).value)

        return values


class ClusterLabelEnum(IntEnum):
    """Enumerates all possible columns of the cluster label tensor."""

    CLUSTER = CLUST_COL
    PARTICLE = PART_COL
    GROUP = GROUP_COL
    ANCESTOR = ANCST_COL
    INTERACTION = INTER_COL
    NU = NU_COL
    PID = PID_COL
    GROUP_PRIMARY = PRGRP_COL
    INTER_PRIMARY = PRINT_COL
    MOMENTUM = MOM_COL


class ShapeEnum(IntEnum):
    """Enumerates all possible shape values."""

    SHOWER = SHOWR_SHP
    TRACK = TRACK_SHP
    MICHEL = MICHL_SHP
    DELTA = DELTA_SHP
    LOWE = LOWES_SHP
    GHOST = GHOST_SHP


class PIDEnum(IntEnum):
    """Enumerates all possible particle species values."""

    PHOTON = PHOT_PID
    ELECTRON = ELEC_PID
    MUON = MUON_PID
    PION = PION_PID
    PROTON = PROT_PID
    KAON = KAON_PID



"""Analysis script used to evaluate the semantic segmentation accuracy."""

import numpy as np
from scipy.special import softmax

from spine.ana.base import AnaBase
from spine.utils.globals import GHOST_SHP, LOWES_SHP, SHAPE_COL

__all__ = ["SegmentAna"]


class SegmentAna(AnaBase):
    """Class which computes and stores the necessary data to build a
    semantic segmentation confusion matrix.
    """

    # Name of the analysis script (as specified in the configuration)
    name = "segment_eval"

    def __init__(
        self,
        summary=True,
        num_classes=GHOST_SHP,
        ghost=False,
        use_fragments=False,
        use_particles=False,
        label_key="seg_label",
        **kwargs,
    ):
        """Initialize the analysis script.

        Parameters
        ----------
        summary : bool, default True
            If `True`, summarize the confusion matrix for each entry. If
            `False`, store one row per pixel in the image (extremely memory
            intensive but gives details about pixel scores).
        num_classes : int, default 5
            Number of pixel classses, excluding the ghost class
        ghost : bool, default False
            Evaluate deghosting performance
        use_fragments : bool, default False
            If `True`, rebuild the segmentation for truth and reco from the
            shape of truth and reco fragments. This method is exact, as long as
            there is no ghost points and the cluster label tensor is untouched.
            If the label tensor is adapted, the original fragment boundaries are
            lost.
        use_particles : bool, default False
            If `True`, rebuild the segmentation for truth and reco from the
            shape of truth and reco particles. This method is imperfect, as the
            shape of showers is determined by the primary shape, which may not
            match the secondary fragment shapes in the original segmentation.
            This method is not compatible with ghost points.
        label_key : str, default 'seg_label'
            Name of the tensor which contains the segmentation labels
        **kwargs : dict, optional
            Additional arguments to pass to :class:`AnaBase`
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Store the basic parameters
        self.summary = summary
        self.num_classes = num_classes
        self.ghost = ghost
        self.use_fragments = use_fragments
        self.use_particles = use_particles
        self.label_key = label_key

        # Basic logic checks
        assert (
            not use_fragments or not use_particles
        ), "Cannot use both fragments and particles."
        assert (
            not (use_fragments or use_particles) or summary
        ), "Cannot store detailed score information from fragments/particles."
        assert (
            not (use_fragments or use_particles) or not ghost
        ), "Cannot produce ghost metrics from fragments/particles."

        # List the necessary data products
        keys = self.keys
        if not use_fragments and not use_particles:
            self.obj_source = None
            keys[label_key] = True
            keys["segmentation"] = True
            if ghost:
                self.num_classes += 1
                keys["ghost"] = True

        else:
            keys["points"] = True
            self.obj_type = "fragments" if use_fragments else "particles"
            for prefix in ["reco", "truth"]:
                keys[f"{prefix}_{self.obj_type}"] = True

        self.keys = keys

        # Initialize the output
        if summary:
            self.initialize_writer("summary")
        else:
            self.initialize_writer("pixel")

    def process(self, data):
        """Store the semantic segmentation metrics for one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Fetch the list of labels and predictions for each pixel in the image
        seg_scores = None
        if not self.use_fragments and not self.use_particles:
            # Get the label/predictions from the raw reconstruction output
            seg_label = data[self.label_key][:, SHAPE_COL].astype(np.int32)
            seg_pred = np.argmax(data["segmentation"], axis=1).astype(np.int32)
            deghost_mask = None
            if self.ghost:
                # If there are ghost, must combine the predictions
                full_seg_pred = np.full_like(seg_label, GHOST_SHP, dtype=np.int32)
                deghost_mask = np.argmax(data["ghost"], axis=1) == 0
                full_seg_pred[deghost_mask] = seg_pred
                seg_pred = full_seg_pred

            if not self.summary:
                # If requested, fetch the individual class softmax scores
                seg_scores = softmax(data["segmentation"], axis=1)
                if deghost_mask is not None:
                    # If there are ghosts, interpret the non-ghost score
                    # as a shared score for all other classes.
                    full_seg_scores = np.zeros(
                        (len(seg_scores), self.num_classes), dtype=np.int32
                    )
                    ghost_scores = softmax(data["ghost"], axis=1)
                    full_seg_scores[:, :-1] = ghost_scores[:0] / (self.num_classes - 1)
                    full_seg_scores[:, -1] = ghost_scores[:, 1]

                    full_seg_scores[deghost_mask, :-1] = seg_scores
                    seg_scores = full_seg_scores

        else:
            # Rebuild the labels/predictions from the fragment/particle objects
            seg_label = np.full(len(data["points"]), LOWES_SHP, dtype=np.int32)
            for obj in data[f"truth_{self.obj_type}"]:
                assert len(obj.index > 0), (
                    "The `index` of true fragments is not filled, indicating "
                    "that the original label tensor was modified. Cannot use "
                    "modified fragments to rebuild semantic labels."
                )
                seg_label[obj.index] = obj.shape

            seg_pred = np.full_like(seg_label, LOWES_SHP)
            for obj in data[f"reco_{self.obj_type}"]:
                seg_pred[obj.index] = obj.shape

        # Store the information
        if not self.summary:
            # Store one row per pixel in the image, including pixel scores
            assert (
                seg_scores is not None
            ), "Segment scores not available for detailed storage."
            for i, (seg_l, seg_p) in enumerate(zip(seg_label, seg_pred)):
                row_dict = {"label": seg_l, "pred": seg_p}
                for s in range(self.num_classes):
                    row_dict[f"score_{s}"] = seg_scores[i, s]

                self.append("pixel", **row_dict)

        else:
            # Store a summary of the confusion per entry (confusion matrix counts)
            count = np.histogram2d(
                seg_pred,
                seg_label,
                bins=[self.num_classes, self.num_classes],
                range=[[0, self.num_classes], [0, self.num_classes]],
            )[0]

            row_dict = {}
            for i in range(self.num_classes):
                for j in range(self.num_classes):
                    row_dict[f"count_{i}{j}"] = int(count[i][j])

            self.append("summary", **row_dict)



from collections import defaultdict

import torch


class DiscriminativeLoss(torch.nn.Module):
    """
    Implementation of the Discriminative Loss Function in Pytorch.
    https://arxiv.org/pdf/1708.02551.pdf
    Note that there are many other implementations in Github, yet here
    we tailor it for use in conjuction with Sparse UResNet.
    """

    def __init__(self, cfg, reduction="sum"):
        super(DiscriminativeLoss, self).__init__()
        self.loss_config = cfg["spice_loss"]
        self.num_classes = self.loss_config.get("num_classes", 5)
        self.depth = self.loss_config.get("stride", 5)

        # Clustering Loss Parameters
        self.loss_hyperparams = {}
        self.loss_hyperparams["intra_weight"] = self.loss_config.get(
            "intra_weight", 1.0
        )
        self.loss_hyperparams["inter_weight"] = self.loss_config.get(
            "inter_weight", 1.0
        )
        self.loss_hyperparams["reg_weight"] = self.loss_config.get("reg_weight", 0.001)
        self.loss_hyperparams["intra_margin"] = self.loss_config.get(
            "intracluster_margin", 0.5
        )
        self.loss_hyperparams["inter_margin"] = self.loss_config.get(
            "intercluster_margin", 1.5
        )

        self.dimension = self.loss_config.get("data_dim", 3)
        self.norm = self.loss_config.get("norm", 2)
        self.use_segmentation = self.loss_config.get("use_segmentation", True)

    def find_cluster_means(self, features, labels):
        """
        For a given image, compute the centroids mu_c for each
        cluster label in the embedding space.
        Inputs:
            features (torch.Tensor): the pixel embeddings, shape=(N, d) where
            N is the number of pixels and d is the embedding space dimension.
            labels (torch.Tensor): ground-truth group labels, shape=(N, )
        Returns:
            cluster_means (torch.Tensor): (n_c, d) tensor where n_c is the number of
            distinct instances. Each row is a (1,d) vector corresponding to
            the coordinates of the i-th centroid.
        """
        clabels = labels.unique(sorted=True)
        cluster_means = []
        for c in clabels:
            index = labels == c
            mu_c = features[index].mean(0)
            cluster_means.append(mu_c)
        cluster_means = torch.stack(cluster_means)
        return cluster_means

    def intra_cluster_loss(self, features, labels, cluster_means, margin=0.5):
        """
        Implementation of variance loss in Discriminative Loss.
        Inputs:
            features (torch.Tensor): pixel embedding, same as in find_cluster_means.
            labels (torch.Tensor): ground truth instance labels
            cluster_means (torch.Tensor): output from find_cluster_means
            margin (float/int): constant used to specify delta_v in paper. Think of it
            as the size of each clusters in embedding space.
        Returns:
            intra_loss: (float) variance loss (see paper).
        """
        intra_loss = 0.0
        n_clusters = len(cluster_means)
        cluster_labels = labels.unique(sorted=True)
        for i, c in enumerate(cluster_labels):
            index = labels == c
            dists = torch.norm(
                features[index] - cluster_means[i] + 1e-8, p=self.norm, dim=1
            )
            hinge = torch.clamp(dists - margin, min=0)
            l = torch.mean(torch.pow(hinge, 2))
            intra_loss += l
        intra_loss /= n_clusters
        return intra_loss

    def inter_cluster_loss(self, cluster_means, margin=1.5):
        """
        Implementation of distance loss in Discriminative Loss.
        Inputs:
            cluster_means (torch.Tensor): output from find_cluster_means
            margin (float/int): the magnitude of the margin delta_d in the paper.
            Think of it as the distance between each separate clusters in
            embedding space.
        Returns:
            inter_loss (float): computed cross-centroid distance loss (see paper).
            Factor of 2 is included for proper normalization.
        """
        inter_loss = 0.0
        n_clusters = len(cluster_means)
        if n_clusters < 2:
            # Inter-cluster loss is zero if there only one instance exists for
            # a semantic label.
            return 0.0
        else:
            for i, c1 in enumerate(cluster_means):
                for j, c2 in enumerate(cluster_means):
                    if i != j:
                        dist = torch.norm(c1 - c2 + 1e-8, p=self.norm)
                        hinge = torch.clamp(2.0 * margin - dist, min=0)
                        inter_loss += torch.pow(hinge, 2)
            inter_loss /= float((n_clusters - 1) * n_clusters)
            return inter_loss

    def regularization(self, cluster_means):
        """
        Implementation of regularization loss in Discriminative Loss
        Inputs:
            cluster_means (torch.Tensor): output from find_cluster_means
        Returns:
            reg_loss (float): computed regularization loss (see paper).
        """
        reg_loss = 0.0
        n_clusters, _ = cluster_means.shape
        for i in range(n_clusters):
            reg_loss += torch.norm(cluster_means[i, :] + 1e-8, p=self.norm)
        reg_loss /= float(n_clusters)
        return reg_loss

    def compute_heuristic_accuracy(self, embedding, truth):
        """
        Compute Adjusted Rand Index Score for given embedding coordinates,
        where predicted cluster labels are obtained from distance to closest
        centroid (computes heuristic accuracy).

        Inputs:
            embedding (torch.Tensor): (N, d) Tensor where 'd' is the embedding dimension.
            truth (torch.Tensor): (N, ) Tensor for the ground truth clustering labels.
        Returns:
            score (float): Computed ARI Score
            clustering (array): the predicted cluster labels.
        """
        from spine.math.metrics import adjusted_rand_score

        nearest = []
        with torch.no_grad():
            cmeans = self.find_cluster_means(embedding, truth)
            for centroid in cmeans:
                dists = torch.sum((embedding - centroid) ** 2, dim=1)
                dists = dists.view(-1, 1)
                nearest.append(dists)
            nearest = torch.cat(nearest, dim=1)
            nearest = torch.argmin(nearest, dim=1)
            pred = nearest.cpu().numpy()
            grd = truth.cpu().numpy()
            score = adjusted_rand_score(pred, grd)
        return score

    def combine(self, features, labels, **kwargs):
        """
        Wrapper function for combining different components of the loss function.
        Inputs:
            features (torch.Tensor): pixel embeddings
            labels (torch.Tensor): ground-truth instance labels
        Returns:
            loss: combined loss, in most cases over a given semantic class.
        """
        # Clustering Loss Hyperparameters
        # We allow changing the parameters at each computation in order
        # to alter the margins at each spatial resolution in multi-scale losses.
        intra_margin = kwargs.get("intra_margin", 0.5)
        inter_margin = kwargs.get("inter_margin", 1.5)
        intra_weight = kwargs.get("intra_weight", 1.0)
        inter_weight = kwargs.get("inter_weight", 1.0)
        reg_weight = kwargs.get("reg_weight", 0.001)

        c_means = self.find_cluster_means(features, labels)
        inter_loss = self.inter_cluster_loss(c_means, margin=inter_margin)
        intra_loss = self.intra_cluster_loss(
            features, labels, c_means, margin=intra_margin
        )
        reg_loss = self.regularization(c_means)

        loss = (
            intra_weight * intra_loss
            + inter_weight * inter_loss
            + reg_weight * reg_loss
        )

        return {
            "loss": loss,
            "intra_loss": intra_weight * float(intra_loss),
            "inter_loss": inter_weight * float(inter_loss),
            "reg_loss": reg_weight * float(reg_loss),
        }

    def combine_multiclass(self, features, slabels, clabels, **kwargs):
        """
        Wrapper function for combining different components of the loss,
        in particular when clustering must be done PER SEMANTIC CLASS.

        NOTE: When there are multiple semantic classes, we compute the DLoss
        by first masking out by each semantic segmentation (ground-truth/prediction)
        and then compute the clustering loss over each masked point cloud.

        INPUTS:
            features (torch.Tensor): pixel embeddings
            slabels (torch.Tensor): semantic labels
            clabels (torch.Tensor): group/instance/cluster labels

        OUTPUT:
            loss_segs (list): list of computed loss values for each semantic class.
            loss[i] = computed DLoss for semantic class <i>.
            acc_segs (list): list of computed clustering accuracy for each semantic class.
        """
        loss, acc_segs = defaultdict(list), defaultdict(float)
        semantic_classes = slabels.unique()
        for sc in semantic_classes:
            index = slabels == sc
            num_clusters = len(clabels[index].unique())
            loss_blob = self.combine(features[index], clabels[index], **kwargs)
            for key, val in loss_blob.items():
                loss[key].append(val)
            # loss['loss'].append(loss_blob['loss'])
            # loss['intra_loss'].append(loss_blob['intra_loss'])
            # loss['inter_loss'].append(loss_blob['inter_loss'])
            # loss['reg_loss'].append(loss_blob['reg_loss'])
            acc = self.compute_heuristic_accuracy(features[index], clabels[index])
            acc_segs["accuracy_{}".format(sc.item())] = acc
        return loss, acc_segs

    def forward(self, out, semantic_labels, group_labels):
        """
        Forward function for the Discriminative Loss Module.

        Inputs:
            out: output of UResNet; embedding-space coordinates.
            semantic_labels: ground-truth semantic labels
            group_labels: ground-truth instance labels
        Returns:
            (dict): A dictionary containing key-value pairs for
            loss, accuracy, etc.
        """
        num_gpus = len(semantic_labels)
        loss = defaultdict(list)
        accuracy = defaultdict(list)

        for i in range(num_gpus):
            slabels = semantic_labels[i][:, -1]
            slabels = slabels.int()
            clabels = group_labels[i][:, -2]
            batch_idx = semantic_labels[i][:, 3]
            embedding = out["cluster_feature"][i]
            nbatch = batch_idx.unique().shape[0]

            for bidx in batch_idx.unique(sorted=True):
                embedding_batch = embedding[batch_idx == bidx]
                slabels_batch = slabels[batch_idx == bidx]
                clabels_batch = clabels[batch_idx == bidx]

                if self.use_segmentation:
                    loss_dict, acc_segs = self.combine_multiclass(
                        embedding_batch,
                        slabels_batch,
                        clabels_batch,
                        **self.loss_hyperparams,
                    )
                    for key, val in loss_dict.items():
                        loss[key].append(sum(val) / len(val))
                    for s, acc in acc_segs.items():
                        accuracy[s].append(acc)
                    acc = sum(acc_segs.values()) / len(acc_segs.values())
                    accuracy["accuracy"].append(acc)
                else:
                    loss["loss"].append(
                        self.combine(
                            embedding_batch, clabels_batch, **self.loss_hyperparams
                        )
                    )
                    acc, _ = self.compute_heuristic_accuracy(
                        embedding_batch, clabels_batch
                    )
                    accuracy["accuracy"].append(acc)

        loss_avg = {}
        acc_avg = defaultdict(float)

        for key, val in loss.items():
            loss_avg[key] = sum(val) / len(val)
        for key, val in accuracy.items():
            acc_avg[key] = sum(val) / len(val)

        res = {}
        res.update(loss_avg)
        res.update(acc_avg)

        return res



"""Trigger information parser post-processor module."""

import os

import pandas as pd

from spine.data import Trigger
from spine.post.base import PostBase

__all__ = ["TriggerProcessor"]


class TriggerProcessor(PostBase):
    """Parses trigger information from a CSV file and adds a new trigger_info
    data product to the data dictionary.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "trigger"

    # Alternative allowed names of the post-processor
    aliases = ("parse_trigger",)

    # Set of data keys needed for this post-processor to operate
    _keys = (("run_info", True),)

    def __init__(
        self, file_path, correct_flash_times=True, flash_keys=None, flash_time_corr_us=4
    ):
        """Initialize the trigger information parser.

        Parameters
        ----------
        file_path : str
            Path to the csv file which contains the trigger information
        correct_flash_times : bool, default True
            If True, corrects the flash times using w.r.t. the trigger times
        flash_keys : List[str], optional
            When correcting flash times, provide the list of flash products
            to correct times for
        flash_time_corr_us : float, default 4
            Systematic correction between the trigger time and the flash time
            in us
        """
        # Initialize the parent class
        super().__init__()

        # Load the trigger information
        if not os.path.isfile(file_path):
            raise FileNotFoundError(f"Cannot find the trigger CSV file: {file_path}")

        self.trigger_dict = pd.read_csv(file_path)

        # Store the parameters
        self.correct_flash_times = correct_flash_times
        if self.correct_flash_times:
            assert flash_keys is not None and len(
                flash_keys
            ), "When correcting flash times, must provide the flash keys."
            self.flash_keys = flash_keys
            self.flash_time_corr_us = flash_time_corr_us

            # Add flash keys to the required data products
            self.update_keys({k: True for k in self.flash_keys})

    def process(self, data):
        """Parse the trigger information of one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Fetch the run info, find the corresponding trigger
        run_info = data["run_info"]
        run_id, event_id = run_info.run, run_info.event
        trigger_mask = (self.trigger_dict["run_number"] == run_id) & (
            self.trigger_dict["event_no"] == event_id
        )
        trigger_info = self.trigger_dict[trigger_mask]

        if len(trigger_info) == 0:
            raise KeyError(
                f"Could not find run {run_id}, event {event_id} in the " "trigger file."
            )
        elif len(trigger_info) > 1:
            raise KeyError(
                f"Found more than one trigger associated with {run_id} "
                f"event {event_id} in the trigger file."
            )

        trigger_info = trigger_info.to_dict(orient="records")[0]

        # Build trigger object
        trigger = Trigger(
            time_s=trigger_info["wr_seconds"],
            time_ns=trigger_info["wr_nanoseconds"],
            beam_time_s=trigger_info["beam_seconds"],
            beam_time_ns=trigger_info["beam_nanoseconds"],
            type=trigger_info.get("trigger_type", -1),
        )

        # If requested, loop over the interaction objects, modify flash times
        if self.correct_flash_times:
            # Loop over flashes, correct the timing (flash times are in us)
            offset = (
                (trigger.time_s - trigger.beam_time_s) * 1e6
                + (trigger.time_ns - trigger.beam_time_ns) * 1e-3
                - self.flash_time_corr_us
            )

            for key in self.flash_keys:
                for flash in data[key]:
                    flash.time += offset

        return {"trigger": trigger}



"""Custom normalization layers."""

import MinkowskiEngine as ME
import torch


class MinkowskiPixelNorm(torch.nn.Module):
    """Pixel Normalization Layer for Sparse Tensors.

    PixelNorm layers were used in NVIDIA's ProGAN.

    This layer normalizes the feature vector in each
    pixel to unit length, and has no trainable weights.

    Original paper: https://arxiv.org/pdf/1710.10196.pdf
    """

    def __init__(self, eps=1e-8):
        """Initialize the normalization layer.

        Parameters
        ----------
        eps : float, default 1e-8
            Ensures non-divergent output features
        """
        # Initialize the parent class
        super(MinkowskiPixelNorm, self).__init__()

        # Store the parameters
        self.eps = eps

    def forward(self, input_data):
        """Pass tensor through the layer.

        Parameters
        ----------
        input_data : ME.SparseTensor
            Sparse input tensor

        Return
        ------
        ME.SparseTensor
            Sparse output tensor
        """
        features = input_data.F
        coords = input_data.C
        norm = torch.sum(torch.pow(features, 2), dim=1, keepdim=True)
        out = features / (norm + self.eps).sqrt()

        return ME.SparseTensor(
            out,
            coordinate_manager=input_data.coordinate_manager,
            coordinate_map_key=input_data.coordinate_map_key,
        )

    def __repr__(self):
        """Representation of the noamlization layer.

        This includes the parameters of the layer.
        """
        suffix = f"({self.num_features}, eps={self.eps})"
        return self.__class__.__name__ + suffix


class MinkowskiAdaIN(torch.nn.Module):
    """Adaptive Instance Normalization Layer.

    Many parts of the code is borrowed from pytorch original
    `BatchNorm` implementation.

    Original paper: https://arxiv.org/pdf/1703.06868.pdf
    """

    def __init__(self, in_channels, eps=1e-5):
        """Initialize the normalization layer.

        Parameters
        ----------
        eps : float, default 1e-5
            Ensures non-divergent output features
        """
        # Initialize the parent class
        super(MinkowskiAdaIN, self).__init__()

        # Store parameters
        self.in_channels = in_channels
        self.eps = eps

        # Initialize weights and biases
        self._weight = torch.ones(in_channels)
        self._bias = torch.zeros(in_channels)

    @property
    def weight(self):
        """Weight parameter of the AdaIN layer.

        Note that in AdaptIS, the parameters to the AdaIN layer
        are trainable outputs from the controller network.
        """
        return self._weight

    @weight.setter
    def weight(self, weight):
        if weight.shape[0] != self.in_channels:
            raise ValueError(
                "Supplied weight vector feature dimension"
                "does not match AdaIN layer definition."
            )
        self._weight = weight

    @property
    def bias(self):
        """Bias parameter of the AdaIN layer.

        Note that in AdaptIS, the parameters to the AdaIN layer
        are trainable outputs from the controller network.
        """
        return self._bias

    @bias.setter
    def bias(self, bias):
        if bias.shape[0] != self.in_channels:
            raise ValueError(
                "Supplied bias vector feature dimension"
                "does not match AdaIN layer definition."
            )
        self._bias = bias

    def forward(self, x):
        """Pass tensor through the layer.

        Parameters
        ----------
        input_data : ME.SparseTensor
            Sparse input tensor

        Return
        ------
        ME.SparseTensor
            Sparse output tensor
        """
        f = x.F
        norm = (f - f.mean(dim=0)) / (f.var(dim=0) + self.eps).sqrt()
        out = self.weight * norm + self.bias

        return ME.SparseTensor(
            out, coords_key=input_data.coords_key, coords_manager=input_data.coords_man
        )



"""Base class for all data representation builders."""

from abc import ABC, abstractmethod

import numpy as np

from spine.data import ObjectList


class BuilderBase(ABC):
    """Abstract base class for building all data structures

    A Builder class takes input data and full chain result dictionaries
    and processes them into human-readable data structures.
    """

    # Builder name
    name = None

    # Types of objects constructed by the builder
    _reco_type = None
    _truth_type = None

    # Necessary/optional data products to build a reconstructed object
    _build_reco_keys = (("points", True), ("depositions", True), ("sources", False))

    # Necessary/optional data products to build a truth object
    _build_truth_keys = (
        ("label_tensor", True),
        ("label_adapt_tensor", False),
        ("label_g4_tensor", False),
        ("points_label", True),
        ("points", False),
        ("points_g4", False),
        ("depositions_label", True),
        ("depositions", False),
        ("depositions_q_label", False),
        ("depositions_g4", False),
        ("sources_label", False),
        ("sources", False),
    )

    # Necessary/optional data products to load a reconstructed object
    _load_reco_keys = (("points", False), ("depositions", False), ("sources", False))

    # Necessary/optional data products to load a truth object
    _load_truth_keys = (
        ("points_label", False),
        ("points", False),
        ("points_g4", False),
        ("depositions_label", False),
        ("depositions", False),
        ("depositions_q_label", False),
        ("depositions_g4", False),
        ("sources_label", False),
        ("sources", False),
    )

    def __init__(self, mode, units):
        """Initializes the builder.

        Parameters
        ----------
        mode : str, default 'both'
            Whether to construct reconstructed objects, true objects or both
            (one of 'reco', 'truth', 'both' or 'all')
        units : str, default 'cm'
            Units in which the position arguments of the constructed objects
            should be expressed (one of 'cm' or 'px')
        """
        # Store the mode and units
        self.mode = mode
        self.units = units

    def __call__(self, data):
        """Build representations for a batch of data.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Dispatch
        for mode, avoid in [("reco", "truth"), ("truth", "reco")]:
            out_key = f"{mode}_{self.name}s"
            if self.mode != avoid:
                if np.isscalar(data["index"]):
                    # Single entry to process
                    data[out_key] = self.process(data, mode)

                else:
                    # Batch of data to process
                    const_list = []
                    for entry in range(len(data["index"])):
                        const_list.append(self.process(data, mode, entry))
                    data[out_key] = const_list

    def process(self, data, mode, entry=None):
        """Build representations for a single entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        mode : str
            Type of object to reconstruct ('reco' or 'truth')
        entry : int, optional
            Entry to process
        """
        # Dispatch to the appropriate function
        key = f"{mode}_{self.name}s"
        if key in data:
            func = f"load_{mode}"
        else:
            func = f"build_{mode}"

        result = self.construct(func, data, entry)

        # When loading, check that the units are as expected
        if "load" in func:
            self.check_units(data, key, entry)

        return result

    def check_units(self, data, key, entry=None):
        """Checks that the objects in the list are expressed in the
        appropriate units. Convert them otherwise.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        key : str
            Dictionary key corresponding to the objects to convert
        entry : int, optional
            Entry to process
        """
        for obj in data[key]:
            if obj.units != self.units:
                assert (
                    "meta" in data
                ), "Cannot convert units without metadata information."
                meta = data["meta"][entry] if entry is not None else data["meta"]
                getattr(obj, f"to_{self.units}")(meta)

    def construct(self, func, data, entry=None):
        """Prepares the input based on the required data and runs constructor.

        Parameters
        ----------
        func : str
            Build function name
        data : dict
            Dictionary of data products
        entry : int, optional
            Entry to process

        Returns
        -------
        List[object]
            List of constructed objects
        """
        # Get the description of the fields needed by this source object
        input_data = {}
        method, dtype = func.split("_")
        keys = getattr(self, f"_{func}_keys")
        for key, req in keys:
            # If the field has no default value, must be provided
            if req and key not in data:
                raise KeyError(
                    f"Must provide `{key}` data product to {method} the "
                    f"{dtype} {self.name}s."
                )

            if key in data:
                if entry is not None:
                    input_data[key] = data[key][entry]
                else:
                    input_data[key] = data[key]

        obj_list = getattr(self, func)(input_data)
        default = getattr(self, f"_{dtype}_type")()

        return ObjectList(obj_list, default)

    @abstractmethod
    def build_reco(self, data):
        """Place-holder for a method used to build reconstructed objects.

        Parameters
        ----------
        data : dict
            Dictionary which contains the necessary data products
        """
        raise NotImplementedError

    @abstractmethod
    def build_truth(self, data):
        """Place-holder for a method used to build truth objects.

        Parameters
        ----------
        data : dict
            Dictionary which contains the necessary data products
        """
        raise NotImplementedError

    @abstractmethod
    def load_reco(self, data):
        """Place-holder for a method used to load reconstructed objects.

        Parameters
        ----------
        data : dict
            Dictionary which contains the necessary data products
        """
        raise NotImplementedError

    @abstractmethod
    def load_truth(self, data):
        """Place-holder for a method used to load truth objects.

        Parameters
        ----------
        data : dict
            Dictionary which contains the necessary data products
        """
        raise NotImplementedError



"""Module that supports TPC-CRT matching."""

import numpy as np

from spine.geo import GeoManager
from spine.math.distance import cdist
from spine.utils.globals import TRACK_SHP


class CRTMatcher:
    """Matches particles and CRT hits by matching the propagated TPC tracks
    with the position of CRT hits.
    """

    # List of valid matching methods
    _match_methods = ("threshold", "best")

    def __init__(
        self,
        driftv,
        match_method="threshold",
        tolerance=5,
        time_window=None,
        min_part_size=None,
        min_crt_pe=None,
        match_distance=None,
    ):
        """Initalize the barycenter flash matcher.

        Parameters
        ----------
        driftv : float
            Drift velocity in cm/us
        match_method : str, default 'distance'
            Matching method (one of 'threshold' or 'best')
            - 'threshold': If a TPC track can be propagated such that it matches
            the CRT hit within some threshold, they are matched
            - 'best': Only allow at most a single TPC match per CRT hit
        tolerance : float, default 5
            Distance along x in cm that a track is allowed to float around the
            expected position given a CRT hit time
        time_window : List, optional
            List of [min, max] values of optical flash times to consider
        first_flash_only : bool, default False
            Only try to match the first flash in the time window
        min_part_size : int, optional
            Minimum number of voxel in an particle to consider it
        min_crt_pe : float, optional
            Minimum number of PE in a CRT hit to consider it
        match_distance : float, optional
            If a threshold is used, specifies the acceptable distance
        """
        # Check validity of key parameters
        if match_method not in self._match_methods:
            raise ValueError(
                f"CRT matching method not recognized: {match_method}. "
                f"Must be one of {self._match_methods}."
            )

        if match_method == "threshold":
            assert match_distance is not None, (
                "When using the `threshold` method, must specify the "
                "`match_distance` parameter."
            )
            self.match_distance = float(match_distance)

        # Store the geometry manager and drift velocity
        self.geo = GeoManager.get_instance()
        self.driftv = driftv

        # Store the flash matching parameters
        self.match_method = match_method
        self.tolerance = tolerance
        self.time_window = time_window
        self.min_part_size = min_part_size
        self.min_crt_pe = min_crt_pe

    def get_matches(self, particles, crthits):
        """Makes [particle, crthit] pairs that are compatible.

        Parameters
        ----------
        particles : List[Union[RecoParticle, TruthParticle]]
            List of particles
        crthits : List[CRTHit]
            List of CRT hits

        Returns
        -------
        List[[Interaction, larcv.Flash, float]]
            List of [particle, flash, distance] triplets
        """
        # Check on the geometry
        assert (
            self.geo.crt is not None
        ), "Cannot find CRT-TPC matches without `crt` geometry."

        # Restrict the CRT hits and particles based on the configuration
        particles, crthits = self.restrict_objects(particles, crthits)
        if len(crthits) == 0 or len(particles) == 0:
            return []

        # For each track, determine how much it is allowed to move along the drift axis
        fit_ranges = np.empty((len(particles), 2), dtype=np.float32)
        for i, part in enumerate(particles):
            # Make sure the sources are provided
            # TODO: change the way points/sources are fetched
            # points = self.get_points(part)
            # sources = self.get_sources(part)
            points, sources = part.points, part.sources
            assert len(sources), "Cannot find CRT-TPC matches without `sources`."

            # Check the TPC composition of the object
            modules, tpcs = self.geo.get_contributors(sources)
            if len(tpcs) == 1:
                # If the particle is made up of a single TPC, it can be
                # placed anywhere within the boundaries of that TPC.
                # Fetch the TPC boundaries along the drift axis
                chamber = self.geo.tpc[modules[0]][tpcs[0]]
                dsign, daxis = chamber.drift_sign, chamber.drift_axis
                apos, cpos = chamber.anode_pos, chamber.cathode_pos
                llim, ulim = (cpos, apos) if dsign > 0 else (apos, cpos)

                # Create boundary conditions
                pmin, pmax = np.min(points[:, daxis]), np.max(points[:, daxis])
                lbound, ubound = llim - pmin, ulim - pmax
                if dsign > 0:
                    fit_ranges[i] = [lbound, ubound]
                else:
                    fit_ranges[i] = [-ubound, -lbound]

            else:
                # If the particle is made up of multiple TPCs, it is a cathode
                # crosser which only has a single allowed position
                fit_ranges[i] = [0.0, 0.0]  # TODO

        # Find matches, start by looping over CRT hits
        matches = []
        for hit in crthits:
            # Fetch the normal to the CRT plane (self.geo.crt guaranteed non-None by __init__ assert)
            crt_plane = self.geo.crt.get_plane(hit.position, hit.plane)
            normal = crt_plane.normal

            # Downselect to particles which can match the time of the hit
            offset = hit.time * self.driftv
            mask = (offset >= fit_ranges[:, 0] - self.tolerance) & (
                offset <= fit_ranges[:, 1] + self.tolerance
            )
            part_ids = np.where(mask)[0]

            # Loop over candidate particles
            for part_id in part_ids:
                # Get the point and direction closest to the CRT hit
                part = particles[part_id]
                points = np.vstack((part.start_point, part.end_point))
                dirs = np.vstack((part.start_dir, part.end_dir))
                closest_id = np.argmin(np.sum((points - hit.center) ** 2, axis=1))
                point, direction = np.copy(points[closest_id]), dirs[closest_id]

                # Move the point according to the expected offset
                # TODO: change the way points/sources are fetched
                closest_id = np.argmin(cdist(part.points, point[None, :]))
                modules, tpcs = self.geo.get_contributors(
                    part.sources[closest_id][None, :]
                )
                chamber = self.geo.tpc[modules[0]][tpcs[0]]
                dsign, daxis = chamber.drift_sign, chamber.drift_axis
                point[daxis] += dsign * offset

                # Check how close they meet in the plane of the CRT
                # TODO: integrate uncertainties properly
                intercept = self.line_plane_intercept(
                    point, direction, hit.center, normal
                )
                dist = np.linalg.norm(intercept - hit.center)
                if dist < self.match_distance:
                    matches.append((part, hit, dist))

        return matches

    def restrict_objects(self, particles, crthits):
        """Restrict the list of particles/CRT hits to match the configuration.

        Parameters
        ----------
        particles : List[Union[RecoParticle, TruthParticle]]
            List of particles
        crthits : List[CRTHit]
            List of CRT hits

        Returns
        -------
        List[Union[RecoParticle, TruthParticle]]
            Restricted list of particles
        List[CRTHit]
            Restricted list of CRT hits
        """
        # Restrict the CRT hits to those that fit the selection criterion
        if self.time_window is not None:
            t1, t2 = self.time_window
            crthits = [h for h in crthits if (h.time > t1 and h.time < t2)]

        # Restrict the particles to tracks that fit the selection criteria
        particles = [part for part in particles if part.shape == TRACK_SHP]
        if self.min_part_size is not None:
            particles = [part for part in particles if part.size > self.min_part_size]

        return particles, crthits

    def line_plane_intercept(self, line_p, line_v, plane_p, plane_n):
        """Find the intercept between a line and a plane in 3D.

        Parameters
        ----------
        line_p : np.ndarray
            (3) Position of a point on the line
        line_v : np.ndarray
            (3) Direction vector of the line
        plane_p : np.ndarray
            (3) Position of a point on the plane
        plane_n : np.ndarray
            (3) Normal to the plane

        Returns
        -------
        np.ndarray
            (3) Position of the intercept
        """
        # Compute dot product, check that line/plane are not parallel
        denom = np.dot(line_v, plane_n)
        if denom == 0.0:
            # Line parallel to plane
            return np.full(3, -np.inf)

        # Compute intercept
        t = np.dot(plane_p - line_p, plane_n) / denom

        return line_p + t * line_v



"""Module that defines a generic node classification loss."""

from warnings import warn

import numpy as np
import torch

from spine.model.layer.factories import loss_fn_factory
from spine.utils.enums import enum_factory
from spine.utils.gnn.cluster import (
    get_cluster_closest_label_batch,
    get_cluster_label_batch,
)
from spine.utils.weighting import get_class_weights

__all__ = ["NodeClassLoss"]


class NodeClassLoss(torch.nn.Module):
    """Generic loss used to train node identification.

    Takes the C-channel node output of the GNN and optimizes node-wise scores
    such that the score corresponding to the correct class is maximized.

    For use in config:

    ..  code-block:: yaml

        model:
          name: grappa
          modules:
            grappa_loss:
              node_loss:
                name: class
                <dictionary of arguments to pass to the loss>

    See configuration files prefixed with `grappa_` under the `config`
    directory for detailed examples of working configurations.
    """

    # Name of the loss (as specified in the configuration)
    name = "class"

    # Alternative allowed names of the loss
    aliases = ("classification",)

    def __init__(
        self,
        target,
        loss="ce",
        balance_loss=False,
        weights=None,
        use_closest=False,
        secondary_label=-1,
    ):
        """Initialize the node classifcation loss function.

        Parameters
        ----------
        target : str
            Column in the label tensor specifying the classification target
        loss : str, default 'ce'
            Name of the loss function to apply
        balance_loss : bool, default False
            Whether to weight the loss to account for class imbalance
        weights : list, optional
            (C) One weight value per class
        use_closest : bool, default False
            For each particle group, assign the label class to the node which
            is closest to the particle start point only
        secondary_label : Union[int, List[int]], default -1
            When using `use_closest=True`, this label is assigned to nodes which
            are not the closest to a the start point of a particle group. These
            numbers can be different for each class if specified as a list
        """
        # Initialize the parent class
        super().__init__()

        # Parse the classification target
        self.target = enum_factory("cluster", target)

        # Initialize basic parameters
        self.balance_loss = balance_loss
        self.weights = weights
        self.use_closest = use_closest
        self.secondary_label = secondary_label

        # Sanity check
        assert (
            weights is None or not balance_loss
        ), "Do not provide weights if they are to be computed on the fly."

        # Set the loss
        self.loss_fn = loss_fn_factory(loss, functional=True)

    def forward(self, clust_label, clusts, node_pred, coord_label=None, **kwargs):
        """Applies the node classification loss to a batch of data.

        Parameters
        ----------
        clust_label : TensorBatch
            (N, 1 + D + N_f) Tensor of cluster labels for the batch
        clusts : IndexBatch
            (C) Index which maps each cluster to a list of voxel IDs
        node_pred : TensorBatch
            (C, 2) Node prediction logits (binary output)
        coord_label : TensorBatch, optional
            (P, 1 + D + 8) Label start, end, time and shape for each point
        **kwargs : dict, optional
            Other labels/outputs of the model which are not relevant here

        Returns
        -------
        loss : torch.Tensor
            Value of the loss
        accuracy : float
            Value of the node-wise classification accuracy
        count : int
            Number of nodes the loss was applied to
        """
        # Get the class labels
        node_assn = get_cluster_label_batch(clust_label, clusts, column=self.target)

        # If requested, adjust the class labeling of particle groups by picking
        # the node closest to the creation point as the reference target
        num_classes = node_pred.shape[1]
        if self.use_closest:
            # Make sure that the start point labeling is provided
            assert coord_label is not None, (
                "To use the node closest to the particle creation point "
                "as the reference node, must provide `coord_label`."
            )

            # Convert the default labels into a list of one value per class
            if np.isscalar(self.secondary_label):
                default = np.full(num_classes, self.secondary_label, dtype=int)
            else:
                assert len(self.secondary_label) == num_classes, (
                    "Must either provide a single default secondary label "
                    "or exactly one per label class."
                )
                default = np.array(self.secondary_label, dtype=int)

            # Adjust the class labels
            node_assn = get_cluster_closest_label_batch(
                clust_label, coord_label, clusts, node_assn, default
            )

        # Create a mask for valid nodes (-1 indicates an invalid class ID)
        valid_mask = node_assn.tensor > -1

        # Check that the labels and the output tensor size are compatible
        class_mask = node_assn.tensor < num_classes
        if np.any(~class_mask):
            warn(
                "There are class labels with a value larger than the "
                f"size of the output logit vector ({num_classes}).",
                RuntimeWarning,
            )

        valid_mask &= class_mask

        # Apply the valid mask and convert the labels to a torch.Tensor
        valid_index = np.where(valid_mask)[0]
        node_assn = node_assn.to_tensor(dtype=torch.long, device=node_pred.device)
        node_assn = node_assn.tensor[valid_index]
        node_pred = node_pred.tensor[valid_index]

        # Compute the loss. Balance classes if requested
        if self.balance_loss:
            self.weights = get_class_weights(node_assn, num_classes=num_classes)

        loss = self.loss_fn(node_pred, node_assn, weight=self.weights, reduction="sum")
        if len(valid_index):
            loss /= len(valid_index)

        # Compute accuracy of assignment (fraction of correctly assigned nodes)
        acc = 1.0
        acc_class = [1.0] * num_classes
        if len(valid_index):
            preds = torch.argmax(node_pred, dim=1)
            acc = float(torch.sum(preds == node_assn))
            acc /= len(valid_index)
            for c in range(num_classes):
                index = torch.where(node_assn == c)[0]
                if len(index):
                    acc_class[c] = float(torch.sum(preds[index] == c)) / len(index)

        # Prepare and return result
        result = {"loss": loss, "accuracy": acc, "count": len(valid_index)}

        for c in range(num_classes):
            result[f"accuracy_class_{c}"] = acc_class[c]

        return result



"""Module that contains all parsers related to LArCV cluster data.

Contains the following parsers:
- :class:`Cluster2DParser`
- :class:`Cluster3DParser`
- :class:`Cluster3DAggregateParser`
- :class:`Cluster3DChargeRescaledParser`
"""

from collections import OrderedDict
from warnings import warn

import numpy as np

from spine.data import Meta
from spine.math.cluster import dbscan
from spine.math.distance import METRICS
from spine.utils.conditional import larcv
from spine.utils.globals import (
    ANCST_COL,
    CLUST_COL,
    DELTA_SHP,
    GROUP_COL,
    INTER_COL,
    NU_COL,
    PART_COL,
    SHAPE_COL,
    SHAPE_PREC,
    VALUE_COL,
)
from spine.utils.particles import process_particle_event
from spine.utils.ppn import image_coordinates

from .base import ParserBase
from .clean_data import clean_sparse_data
from .data import ParserTensor
from .sparse import (
    Sparse3DAggregateParser,
    Sparse3DChargeRescaledParser,
    Sparse3DParser,
)

__all__ = [
    "Cluster2DParser",
    "Cluster3DParser",
    "Cluster3DAggregateParser",
    "Cluster3DChargeRescaledParser",
]


class Cluster2DParser(ParserBase):
    """Class that retrieves and parses a 2D cluster list.

    .. code-block. yaml

        schema:
          cluster_label:
            parser: cluster2d
            cluster_event: cluster2d_pcluster
    """

    # Name of the parser (as specified in the configuration)
    name = "cluster2d"

    # Type of object(s) returned by the parser
    returns = "tensor"

    def __init__(self, dtype, cluster_event, projection_id):
        """Initialize the parser.

        Parameters
        ----------
        cluster_event : larcv.EventClusterPixel2D
            Event which contains the 2D clusters
        projection_id : int
            Projection ID to get the 2D images from
        """
        # Initialize the parent class
        super().__init__(dtype, cluster_event=cluster_event)

        # Store the revelant attributes
        self.projection_id = projection_id

        # Define the overlay strategy parameters
        self.index_cols = np.array([CLUST_COL])
        self.sum_cols = np.array([VALUE_COL])

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, cluster_event):
        """Converts a 2D clusters tensor into a single tensor.

        Parameters
        ----------
        cluster_event : larcv.EventClusterPixel2D
            Event which contains the 2D clusters

        Returns
        -------
        ParserTensor
            coords : np.ndarray
                (N, 2) array of [x, y] coordinates
            features : np.ndarray
                (N, 2) array of [pixel value, cluster ID]
            meta : Meta
                Metadata of the parsed image
        """
        # Get the cluster from the appropriate projection
        cluster_event_p = cluster_event.cluster_pixel_2d(self.projection_id)

        # Loop over clusters, store information
        meta = cluster_event_p.meta()
        num_clusters = cluster_event_p.size()
        clusters_voxels, clusters_features = [], []
        for i in range(num_clusters):
            cluster = cluster_event_p.as_vector()[i]
            num_points = cluster.as_vector().size()
            if num_points > 0:
                x = np.empty(num_points, dtype=np.int32)
                y = np.empty(num_points, dtype=np.int32)
                value = np.empty(num_points, dtype=np.float32)
                larcv.as_flat_arrays(cluster, meta, x, y, value)
                pixels = np.stack([x, y], axis=1).astype(self.itype)
                value = value.astype(self.ftype)
                cluster_id = np.full(num_points, i, dtype=self.ftype)
                clusters_voxels.append(pixels)
                clusters_features.append(np.column_stack([value, cluster_id]))

        # Concatenate clusters
        if not clusters_voxels:
            np_voxels = np.empty((0, 2), dtype=self.ftype)
            np_features = np.empty((0, 2), dtype=self.ftype)
        else:
            np_voxels = np.concatenate(clusters_voxels, axis=0)
            np_features = np.concatenate(clusters_features, axis=0)

        # Evaluate shifts to apply to each index column
        index_shifts = np.max(np_features[:, -1], keepdims=True, initial=-1) + 1

        return ParserTensor(
            coords=np_voxels,
            features=np_features,
            meta=Meta.from_larcv(meta),
            remove_duplicates=True,
            index_shifts=index_shifts,
            index_cols=self.index_cols,
            sum_cols=self.sum_cols,
        )


class Cluster3DParser(ParserBase):
    """Class that retrieves and parses a 3D cluster list.

    .. code-block. yaml

        schema:
          cluster_label:
            parser: cluster3d
            cluster_event: cluster3d_pcluster
            particle_event: particle_pcluster
            particle_mpv_event: particle_mpv
            neutrino_event: neutrino_mpv
            sparse_semantics_event: sparse3d_semantics
            sparse_value_event: sparse3d_pcluster
            add_particle_info: true
            clean_data: true
            type_include_mpr: false
            type_include_secondary: false
            primary_include_mpr: true
            break_clusters: false
    """

    # Name of the parser (as specified in the configuration)
    name = "cluster3d"

    # Type of object(s) returned by the parser
    returns = "tensor"

    def __init__(
        self,
        dtype,
        particle_event=None,
        add_particle_info=False,
        clean_data=False,
        type_include_mpr=True,
        type_include_secondary=True,
        primary_include_mpr=True,
        break_clusters=False,
        break_eps=1.1,
        break_metric="chebyshev",
        shape_precedence=SHAPE_PREC,
        **kwargs,
    ):
        """Initialize the parser.

        Parameters
        ----------
        particle_event : larcv.EventParticle, optional
            List of true particle information. If prodided, allows to fetch
            more information about each of the pixels in the image
        add_particle_info : bool, default False
            If `True`, adds truth information from the true particle list
        clean_data : bool, default False
            If `True`, removes duplicate voxels
        type_include_mpr : bool, default False
            If `False`, sets all PID labels to -1 for MPR particles
        type_include_secondary : bool, default False
            If `False`, sets all PID labels to -1 for secondary particles
        primary_include_mpr : bool, default False
            If `False`, sets all primary labels to -1 for MPR particles
        break_clusters : bool, default False
            If `True`, runs DBSCAN on each cluster, assigns different cluster
            IDs to different fragments of the broken-down cluster
        break_eps : float, default 1.1
            Distance scale used in the break up procedure
        break_metric : str, default 'chebyshev'
            Distance metric used in the break up produce
        shape_precedence: list, default SHAPE_PREC
             Array of classes in the reference array, ordered by precedence
        **kwargs : dict, optional
            Data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(dtype, particle_event=particle_event, **kwargs)

        # Store the revelant attributes
        self.add_particle_info = add_particle_info
        self.clean_data = clean_data
        self.type_include_mpr = type_include_mpr
        self.type_include_secondary = type_include_secondary
        self.primary_include_mpr = primary_include_mpr
        self.shape_precedence = shape_precedence

        # Intialize DBSCAN if the clusters are to be broken up
        self.break_clusters = break_clusters
        self.break_eps = break_eps
        self.break_metric_id = METRICS[break_metric]

        # Intialize the sparse and particle parsers
        self.sparse_parser = Sparse3DParser(dtype, sparse_event="dummy")

        # If particle information is to be incldued, check that it is provided
        if self.add_particle_info:
            # Must provide particle event to build particle info
            assert particle_event is not None, (
                "If `add_particle_info` is `True`, must provide the "
                "`particle_event` argument."
            )

        # Define the overlay strategy parameters
        self.sum_cols = np.array([VALUE_COL])
        if not self.add_particle_info:
            self.index_cols = np.array([CLUST_COL])
            self.prec_col = None
        else:
            self.index_cols = np.array(
                [CLUST_COL, PART_COL, GROUP_COL, ANCST_COL, INTER_COL, NU_COL]
            )
            self.prec_col = SHAPE_COL

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(
        self,
        cluster_event,
        particle_event=None,
        particle_mpv_event=None,
        neutrino_event=None,
        sparse_semantics_event=None,
        sparse_value_event=None,
    ):
        """Parse a list of 3D clusters into a single tensor.

        Parameters
        ----------
        cluster_event : larcv.EventClusterVoxel3D
            Event which contains the 3D clusters
        particle_event : larcv.EventParticle, optional
            List of true particle information. If prodided, allows to fetch
            more information about each of the pixels in the image
        particle_mpv_event : larcv.EventParticle, optional
            List of true particle information for MPV particles only. If
            provided, it is used to determine which particles are MPV
        particle_mpv_event: larcv.EventNeutrino, optional
            List of true neutrino information. If provided, it is used
            to determine which particles are MPV
        sparse_semantics_event : larcv.EventSparseTensor3D, optional
            Semantics of each of the voxels in the image. If provided,
            overrides the order of precedence used in combining clusters
            which share voxels.
        sparse_value_event : larcv.EventSparseTensor3D, optional
            Value of each of the voxels in the image. If provided,
            overrides the value provided byt eh list of 3D clusters itself

        Returns
        -------
        ParserTensor
            coords : np.ndarray
                (N, 3) array of [x, y, z] coordinates
            features : np.ndarray
                (N, 2/14) array of features, minimally [voxel value, cluster ID].
                If `add_particle_info` is `True`, the additonal columns are
                [particle ID, group ID, interaction ID, neutrino ID, particle type,
                group primary bool, interaction primary bool, vertex x, vertex y,
                vertex z, momentum, semantic type]
            meta : Meta
                Metadata of the parsed image
        """
        # Get the cluster-wise information first
        meta = cluster_event.meta()
        num_clusters = cluster_event.as_vector().size()
        labels = OrderedDict()
        labels["cluster"] = np.arange(num_clusters)
        num_particles = num_clusters
        if self.add_particle_info:
            # Check that that particle objects are of the expected length
            num_particles = particle_event.size()
            assert num_particles in (num_clusters, num_clusters - 1), (
                f"The number of particles ({num_particles}) must be "
                f"aligned with the number of clusters ({num_clusters}). "
                f"There can me one more catch-all cluster at the end."
            )

            # Load up the particle list
            particles = list(particle_event.as_vector())

            # Fetch the variables missing from the larcv objects
            (
                _,
                group_ids,
                ancestor_ids,
                interaction_ids,
                nu_ids,
                group_primaries,
                inter_primaries,
                types,
            ) = process_particle_event(
                particle_event, particle_mpv_event, neutrino_event
            )

            # Store the cluster ID information
            labels["cluster"] = [p.id() for p in particles]
            labels["part"] = [p.id() for p in particles]
            labels["group"] = group_ids
            labels["ancst"] = ancestor_ids
            labels["inter"] = interaction_ids
            labels["nu"] = nu_ids

            # Store the type/primary status
            labels["type"] = types
            labels["pgroup"] = group_primaries
            labels["pinter"] = inter_primaries

            # Store the vertex and momentum
            anc_pos = np.empty((len(particles), 3), dtype=self.ftype)
            for i, p in enumerate(particles):
                anc_pos[i] = image_coordinates(meta, p.ancestor_position())
            labels["vtx_x"] = anc_pos[:, 0]
            labels["vtx_y"] = anc_pos[:, 1]
            labels["vtx_z"] = anc_pos[:, 2]
            labels["p"] = [p.p() for p in particles]

            # Store the shape last (consistent with semantics tensor)
            labels["shape"] = [p.shape() for p in particles]

            # If requested, give invalid labels to a subset of particles
            if not self.type_include_secondary:
                secondary_mask = np.where(np.array(labels["pinter"]) < 1)[0]
                labels["type"] = np.asarray(labels["type"])
                labels["type"][secondary_mask] = -1

            if not self.type_include_mpr or not self.primary_include_mpr:
                mpr_mask = np.where(np.array(labels["nu"]) < 0)[0]
                if not self.type_include_mpr:
                    labels["type"] = np.asarray(labels["type"])
                    labels["type"][mpr_mask] = -1
                if not self.primary_include_mpr:
                    labels["pinter"] = np.asarray(labels["pinter"])
                    labels["pinter"][mpr_mask] = -1

            # Count the number of neutrinos
            if neutrino_event is not None:
                num_neutrinos = len(neutrino_event.as_vector())
            else:
                num_neutrinos = np.max(nu_ids, initial=-1) + 1

        # Loop over clusters, store information
        clusters_voxels, clusters_features = [], []
        id_offset = 0
        for i in range(num_clusters):
            cluster = cluster_event.as_vector()[i]
            num_points = cluster.as_vector().size()
            if num_points > 0:
                # Get the position and pixel value from EventSparseTensor3D
                x = np.empty(num_points, dtype=np.int32)
                y = np.empty(num_points, dtype=np.int32)
                z = np.empty(num_points, dtype=np.int32)
                value = np.empty(num_points, dtype=np.float32)
                larcv.as_flat_arrays(cluster, meta, x, y, z, value)
                voxels = np.stack([x, y, z], axis=1).astype(self.itype)
                value = value.astype(self.ftype)
                clusters_voxels.append(voxels)

                # Append the cluster-wise information
                features = [value]
                for l in labels.values():
                    val = l[i] if i < num_particles else -1
                    features.append(np.full(num_points, val, dtype=self.ftype))

                # If requested, break cluster into detached pieces
                if self.break_clusters:
                    frag_labels = dbscan(
                        voxels,
                        eps=self.break_eps,
                        min_samples=1,
                        metric_id=self.break_metric_id,
                    )
                    features[1] = id_offset + frag_labels
                    id_offset += np.max(frag_labels, initial=-1) + 1

                clusters_features.append(np.column_stack(features))

        # If there are no non-empty clusters, return. Concatenate otherwise
        if not clusters_voxels:
            np_voxels = np.empty((0, 3), dtype=self.itype)
            np_features = np.empty((0, len(labels) + 1), dtype=self.ftype)
        else:
            np_voxels = np.concatenate(clusters_voxels, axis=0)
            np_features = np.concatenate(clusters_features, axis=0)

        # If requested, remove duplicate voxels (cluster overlaps) and
        # match the semantics to those of the provided reference
        if len(np_voxels) and (
            (sparse_semantics_event is not None) or (sparse_value_event is not None)
        ):
            if not self.clean_data:
                warn(
                    "You must set `clean_data` to `True` if you specify a "
                    "sparse tensor in `Cluster3DParser`."
                )
                self.clean_data = True

            # Extract voxels and features
            assert self.add_particle_info, (
                "Need to add particle info to fetch particle "
                "semantics for each voxel."
            )
            assert (
                sparse_semantics_event is not None
            ), "Need to provide a semantics tensor to clean up output."
            tensor_seg = self.sparse_parser.process(sparse_semantics_event)
            np_voxels, np_features = clean_sparse_data(
                np_voxels,
                np_features,
                tensor_seg.coords,
                precedence=self.shape_precedence,
                sum_cols=self.sum_cols - VALUE_COL,
            )

            # Match the semantic column to the reference tensor
            np_features[:, -1] = tensor_seg.features[:, -1]

            # Set all cluster labels to -1 if semantic class is LE or ghost
            shape_mask = tensor_seg.features[:, -1] > DELTA_SHP
            np_features[shape_mask, 1:-1] = -1

            # If a value tree is provided, override value colum
            if sparse_value_event:
                tensor_val = self.sparse_parser.process(sparse_value_event)
                np_features[:, 0] = tensor_val.features[:, 0]

        # Evaluate shifts to apply to each index column
        clust_shift = np.max(np_features[:, 1], initial=-1) + 1
        if not self.add_particle_info:
            index_shifts = np.array([clust_shift])
        else:
            cs, ps, ns = clust_shift, num_particles, num_neutrinos
            index_shifts = np.array([cs, ps, ps, ps, ns])

        return ParserTensor(
            coords=np_voxels,
            features=np_features,
            meta=Meta.from_larcv(meta),
            remove_duplicates=True,
            index_shifts=index_shifts,
            index_cols=self.index_cols,
            sum_cols=self.sum_cols,
            prec_col=self.prec_col,
            precedence=self.shape_precedence,
        )


class Cluster3DAggregateParser(Cluster3DParser):
    """Identical to :class:`Cluster3DParser`, but aggregates charge information
    from multiple value sources.
    """

    # Name of the parser (as specified in the configuration)
    name = "cluster3d_aggr"

    def __init__(self, dtype, sparse_value_event_list, value_aggr, **kwargs):
        """Initialize the parser.

        Parameters
        ----------
        sparse_value_event_list : List[larcv.EventSparseTensor3D]
            List of sparse tensors used to compute the aggregated charge
        value_aggr : str
            Value aggregation function to apply ('sum', 'mean', 'max', etc.)
        **kwargs : dict, optional
            Data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(
            dtype, sparse_value_event_list=sparse_value_event_list, **kwargs
        )

        # Initialize the sparse parser which computes the rescaled charge
        self.sparse_aggr_parser = Sparse3DAggregateParser(
            dtype, sparse_event_list=sparse_value_event_list, aggr=value_aggr
        )

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process_aggr(**self.get_input_data(trees))

    def process_aggr(self, sparse_value_event_list, **kwargs):
        """Parse a list of 3D clusters into a single tensor and fetch the
        value column by aggregating multiple tensor features.

        Parameters
        ----------
        sparse_value_event_list : List[larcv.EventSparseTensor3D]
            List of sparse value tensors
        **kwargs : dict, optional
            Extra data products to pass to the parent Cluster3DParser

        Returns
        -------
        ParserTensor
            coords : np.ndarray
                (N, 3) array of [x, y, z] coordinates
            features : np.ndarray
                (N, 2/14) array of features, minimally [voxel value, cluster ID].
                If `add_particle_info` is `True`, the additonal columns are
                [group ID, interaction ID, neutrino ID, particle type,
                group primary bool, interaction primary bool, vertex x, vertex y,
                vertex z, momentum, semantic type, particle ID]
            meta : Meta
                Metadata of the parsed image
        """
        # Process the input using the main parser
        tensor = self.process(**kwargs)

        # Modify the value column using the aggregate tensor values
        tensor_val = self.sparse_aggr_parser.process_aggr(sparse_value_event_list)
        tensor.features[:, 0] = tensor_val.features[:, 0]

        return tensor


class Cluster3DChargeRescaledParser(Cluster3DParser):
    """Identical to :class:`Cluster3DParser`, but computes rescaled charges
    on the fly.
    """

    # Name of the parser (as specified in the configuration)
    name = "cluster3d_rescale_charge"

    def __init__(
        self,
        dtype,
        sparse_value_event_list,
        collection_only=False,
        collection_id=2,
        **kwargs,
    ):
        """Initialize the parser.

        Parameters
        ----------
        sparse_value_event_list : List[larcv.EventSparseTensor3D]
            (7) List of sparse tensors used to compute the rescaled charge
            - Charge value of each of the contributing planes (3)
            - Index of the plane hit contributing to the space point (3)
            - Semantic labels (1)
        collection_only : bool, default False
            If True, only uses the collection plane charge
        collection_id : int, default 2
            Index of the collection plane
        **kwargs : dict, optional
            Data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(
            dtype, sparse_value_event_list=sparse_value_event_list, **kwargs
        )

        # Initialize the sparse parser which computes the rescaled charge
        self.sparse_rescale_parser = Sparse3DChargeRescaledParser(
            dtype,
            sparse_event_list=sparse_value_event_list,
            collection_only=collection_only,
            collection_id=collection_id,
        )

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process_rescale(**self.get_input_data(trees))

    def process_rescale(self, sparse_value_event_list, **kwargs):
        """Parse a list of 3D clusters into a single tensor and reset
        the value column by rescaling the charge coming from 3 wire planes.

        Parameters
        ----------
        sparse_value_event_list : List[larcv.EventSparseTensor3D]
            (7) List of sparse tensors used to compute the rescaled charge
            - Charge value of each of the contributing planes (3)
            - Index of the plane hit contributing to the space point (3)
            - Semantic labels (1)
        **kwargs : dict, optional
            Extra data products to pass to the parent Cluster3DParser

        Returns
        -------
        ParserTensor
            coords : np.ndarray
                (N, 3) array of [x, y, z] coordinates
            features : np.ndarray
                (N, 2/14) array of features, minimally [voxel value, cluster ID].
                If `add_particle_info` is `True`, the additonal columns are
                [group ID, interaction ID, neutrino ID, particle type,
                group primary bool, interaction primary bool, vertex x, vertex y,
                vertex z, momentum, semantic type, particle ID]
            meta : Meta
                Metadata of the parsed image
        """
        # Process the input using the main parser
        tensor = self.process(**kwargs)

        # Modify the value column using the charge rescaled on the fly
        tensor_val = self.sparse_rescale_parser.process_rescale(sparse_value_event_list)
        tensor.features[:, 0] = tensor_val.features[:, 0]

        return tensor



"""Numba JIT compiled implementation of neighbor query routines.

In particular, this module supports:
- Radius-based neighbor classification
- kNN-based neighbor classification
"""

import numba as nb
import numpy as np

from .base import mode
from .distance import METRICS, cdist, get_metric_id

__all__ = ["RadiusNeighborsClassifier", "KNeighborsClassifier"]


RNC_DTYPE = (
    ("radius", nb.float32),
    ("metric_id", nb.int64),
    ("p", nb.float32),
    ("iterate", nb.boolean),
)


KNC_DTYPE = (("k", nb.int64), ("metric_id", nb.int64), ("p", nb.float32))


@nb.experimental.jitclass(RNC_DTYPE)
class RadiusNeighborsClassifier:
    """Class which assigns labels to points based on radial neighborhood
    majority vote.

    More specifically, for each point that is to be labeled:
    - Find all labeled points within some radius R;
    - Label the point based on majority vote.

    If there are no labeled points in the neighborhood of a query point, a
    label of -1 is assigned to the query point.

    Currently this is bruteforced with cdist, but in the future this is
    intended to be used with a KDTree backend for quicker query.

    Attributes
    ----------
    radius : float
        Radius around which to check
    metric_id : int
        Distance metric enumerator
    p : float
        p-norm factor for the Minkowski metric, if used
    iterate : bool
        Whether to recurse the search until no new labels are assigned
    """

    def __init__(
        self,
        radius: nb.float32,
        metric: nb.types.string = "euclidean",
        p: nb.float32 = 2.0,
        iterate: nb.boolean = True,
    ):
        """Initialize the RadiusNeighborsClassifier parameters.

        Parameters
        ----------
        radius : float
            Radius around which to check
        metric : str, default 'euclidean'
            Distance metric
        p : float, default 2.
            p-norm factor for the Minkowski metric, if used
        iterate : bool, default True
            Whether to recurse the search until no new labels are assigned
        """
        # For Euclidean, save time by using squared Euclidean
        if metric == "euclidean":
            metric = "sqeuclidean"
            radius = radius * radius

        # Store parameters
        self.radius = radius
        self.metric_id = get_metric_id(metric, p)
        self.p = p
        self.iterate = iterate

    def fit_predict(self, X: nb.float32[:, :], y: nb.float32[:], Xq: nb.float32[:, :]):
        """Assign labels to a set of points given a set of reference points.

        Parameters
        ----------
        X : np.ndarray
            (N, 3) Set of reference points
        y : np.ndarray
            (N) Labels of reference points
        Xq : nb.ndarray
            (M, 3) Set of query points

        Returns
        -------
        np.ndarray
            (M) Labels assigned to the query points
        np.ndarray
            Index of points which have not been sucessfully assigned
        """
        # Loop over query points until no new labels can be assigned
        num_query = len(Xq)
        labels = np.empty(num_query, dtype=np.int64)
        orphan_index = np.arange(num_query, dtype=np.int64)
        while num_query > 0:
            # Start by computing the distance between the query and reference
            dists = cdist(Xq, X, metric_id=self.metric_id, p=self.p)

            # Fetch the mask of reference points closer than some radius
            mask = dists < self.radius

            # Loop over query points
            assigned = np.zeros(num_query, dtype=nb.boolean)
            for i in range(num_query):
                # Find the set of points within the predefined radius
                index = np.where(mask[i])[0]

                # Use the mode to define the label
                if len(index):
                    labels[orphan_index[i]] = mode(y[index])
                    assigned[i] = True
                else:
                    labels[orphan_index[i]] = -1

            # If the number of orphans is unchanged, break
            orphan_update = np.where(~assigned)[0]
            if len(orphan_update) == 0 or len(orphan_update) == num_query:
                orphan_index = orphan_index[orphan_update]
                break

            # If no recursion is required, abort loop
            if not self.iterate:
                orphan_index = orphan_index[orphan_update]
                break

            # Update the reference and query points
            label_update = np.where(assigned)[0]
            X = Xq[label_update]
            Xq = Xq[orphan_update]
            y = labels[orphan_index[label_update]]

            # Update orphan list
            orphan_index = orphan_index[orphan_update]
            num_query = len(orphan_index)

        return labels, orphan_index


@nb.experimental.jitclass(KNC_DTYPE)
class KNeighborsClassifier:
    """Class which assigns labels to points based on a nearest neighbor
    majority vote.

    More specifically, for each point that is to be labeled:
    - Find the k closest labeled points;
    - Label the point based on majority vote.

    If there are no labeled points in the neighborhood of a query point, a
    label of -1 is assigned to the query point.

    Currently this is bruteforced with cdist, but in the future this is
    intended to be used with a KDTree backend for quicker query.

    Attributes
    ----------
    k : int
        Number of neighbors to query
    metric_id : int
        Distance metric enumerator
    p : float
        p-norm factor for the Minkowski metric, if used
    """

    def __init__(
        self, k: nb.int64, metric: nb.types.string = "euclidean", p: nb.float32 = 2.0
    ):
        """Initialize the RadiusNeighborsClassifier parameters.

        Parameters
        ----------
        k : int
            Number of neighbors to query
        metric : str, default 'euclidean'
            Distance metric
        p : float, default 2.
            p-norm factor for the Minkowski metric, if used
        """
        # For Euclidean, save time by using squared Euclidean
        if metric == "euclidean":
            metric = "sqeuclidean"

        # Store parameters
        self.k = k
        self.metric_id = get_metric_id(metric, p)
        self.p = p

    def fit_predict(self, X: nb.float32[:, :], y: nb.float32[:], Xq: nb.float32[:, :]):
        """Assign labels to a set of points given a set of reference points.

        Parameters
        ----------
        X : np.ndarray
            (N, 3) Set of reference points
        y : np.ndarray
            (N) Labels of reference points
        Xq : nb.ndarray
            (M, 3) Set of query points

        Returns
        -------
        np.ndarray
            (M) Labels assigned to the query points
        np.ndarray
            Index of points which have not been sucessfully assigned
        """
        # If there are no labeled points provided, nothing to do
        if len(X) == 0:
            return (
                np.full(len(Xq), -1, dtype=np.int64),
                np.arange(len(Xq), dtype=np.int64),
            )

        # Start by computing the distance between the query and reference
        dists = cdist(Xq, X, metric_id=self.metric_id, p=self.p)

        # Loop over query poins
        labels = np.empty(len(Xq), dtype=np.int64)
        for i in range(len(Xq)):
            # Find the list k closest labels
            index = np.argsort(dists[i])[: self.k]

            # Use the mode to define the label
            labels[i] = mode(y[index])

        return labels, np.empty(0, dtype=np.int64)



"""PPN point construction module."""

import numpy as np
from scipy.spatial.distance import cdist

from spine.post.base import PostBase
from spine.utils.globals import COORD_COLS, PPN_SHAPE_COL
from spine.utils.ppn import PPNPredictor

__all__ = ["PPNProcessor"]


class PPNProcessor(PostBase):
    """Run the PPN post-processing function to produce PPN candidate points
    from the raw PPN output.

    If requested, for each particle, match ppn_points that have hausdorff
    distance less than a threshold and update the particle `ppn_candidates`
    attribute in place.

    If `restrict_shape` is `True`, points will be matched to particles with
    the same predicted semantic type only.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "ppn"

    # Alternative allowed names of the post-processor
    aliases = ("get_ppn_candidates",)

    # Set of data keys needed for this post-processor to operate
    _keys = (
        ("segmentation", True),
        ("ppn_points", True),
        ("ppn_coords", True),
        ("ppn_masks", True),
        ("ppn_classify_endpoints", False),
    )

    def __init__(
        self,
        assign_to_particles=False,
        restrict_shape=False,
        match_threshold=2.0,
        **ppn_pred_cfg,
    ):
        """Store the `get_ppn_predictions` keyword arguments.

        Parameters
        ----------
        assign_to_particles: bool, default False
            If `True`, will assign PPN candidates to particle objects
        restrict_shape : bool, default False
            If `True`, only associate PPN candidates with compatible shape
        match_threshold : float, default 2.
            Maximum distance required to assign ppn point to particle
        **ppn_pred_cfg : dict, optional
            Keyword arguments to pass to the `PPNPredictor` class

        Returns
        -------
        dict
            Update result dictionary containing 'ppn_candidates' key
        """
        # Intialize the parent class
        obj_type = "particle" if assign_to_particles else None
        super().__init__(obj_type, "reco")

        # Store the relevant parameters
        self.assign_to_particles = assign_to_particles
        self.restrict_shape = restrict_shape
        self.match_threshold = match_threshold
        self.ppn_predictor = PPNPredictor(**ppn_pred_cfg)

    def process(self, data):
        """Produce PPN candidates for one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Get the PPN candidates
        # TODO: remove requirement to nest output
        data_nest = {k: [v] for k, v in data.items()}
        ppn_pred = self.ppn_predictor(**data_nest)[0]
        result = {"ppn_pred": ppn_pred}

        # If requested, assign PPN candidates to particles
        if self.assign_to_particles:
            ppn_points = ppn_pred[:, COORD_COLS]
            for part in data["reco_particles"]:
                # Get the valid list of candidates
                valid_index = np.arange(len(ppn_pred))
                if not self.restrict_shape:
                    candidates = ppn_points
                else:
                    valid_index = np.where(ppn_pred[:, PPN_SHAPE_COL] == part.shape)[0]
                    candidates = ppn_points[valid_index]

                # Restrict to points that are susfficienly close
                dists = np.min(cdist(candidates, part.points), axis=1)
                dist_index = np.where(dists < self.match_threshold)[0]

                # Compute and sotre points matches to this particle
                matches = candidates[dist_index]
                part.ppn_points = matches
                if not self.restrict_shape:
                    part.ppn_ids = dist_index
                else:
                    part.ppn_ids = valid_index[dist_index]

        return result



"""Contains functions needed to instantiate a class from a dictionary.

This allows to generically convert a YAML block into an instatiated class
with all the appropriate checks that the class exists and is provided
with appropriate arguments.
"""

from copy import deepcopy
from warnings import warn

from .logger import logger


def module_dict(module, class_name=None, pattern=None):
    """Converts module into a dictionary which maps class names onto classes.

    Parameters
    ----------
    module : module
        Module from which to fetch the classes
    class_name : str, optional
        If specified, only allow aliases that match it
    pattern : str, optional
        If specified, looks for a specific pattern in the class name

    Returns
    -------
    dict
        Dictionary which maps acceptable class names to classes themselves
    """
    # Loop over classes/functions in the module
    module_dict = {}
    cls_names = getattr(module, "__attr__", dir(module))
    for cls_name in cls_names:
        # Skip private objects
        if cls_name[0] == "_":
            continue

        # If a pattern is specified, check for it in the class name
        cls = getattr(module, cls_name)
        if pattern is not None and pattern not in cls.__name__:
            continue

        # Only consider classes which belong to the module of interest
        if hasattr(cls, "__module__") and module.__name__ in cls.__module__:

            # Store the class name as an option to fetch it
            module_dict[cls_name] = cls

            # If a name is provided, add it to the allowed options
            if hasattr(cls, "name") and len(cls.name):
                module_dict[cls.name] = cls

            # If aliases are specified, it is allowed but should be avoided
            if hasattr(cls, "aliases"):
                for al in cls.aliases:
                    if class_name is not None and class_name == al:
                        warn(
                            f"This name ({al}) is deprecated. Use "
                            f"{cls.name} instead.",
                            DeprecationWarning,
                        )
                        module_dict[al] = cls
                    else:
                        module_dict[al] = cls

    return module_dict


def instantiate(module_dict, cfg, alt_name=None, **kwargs):
    """Instantiates a class based on a configuration dictionary and a list of
    possible classes to chose from.

    This function supports two YAML configuration structures
    (parsed as a dictionary):

    .. code-block:: yaml

        function:
          name: function_name
          kwarg_1: value_1
          kwarg_2: value_2
          ...

    or

    .. code-block:: yaml

        function:
          name: function_name
          args:
            value_1
            value_2
            ...
          kwargs:
            kwarg_1: value_1
            kwarg_2: value_2
            ...

    The `name` field can have a different name, as long as it is specified.

    Parameters
    ----------
    module_dict : dict
        Dictionary which maps a class name onto an object class.
    cfg : dict
        Configuration dictionary
    alt_name : str, optional
        Key under which the class name can be specfied, beside 'name' itself
    **kwargs : dict, optional
        Additional parameters to pass to the function

    Returns
    -------
    object
        Instantiated object
    """
    # If the configuration is a string, assume it is a class name with no
    # parameters to be passed to it
    if isinstance(cfg, str):
        cfg = {"name": cfg}

    # Get the name of the class, check that it exists
    config = deepcopy(cfg)
    if alt_name is not None:
        assert (alt_name in config) ^ (
            "name" in config
        ), f"Should specify one of `name` or `{alt_name}`"
        name = alt_name if alt_name in config else "name"
    else:
        assert "name" in config, "Could not find the name of the class under `name`"
        name = "name"

    class_name = config.pop(name)

    # Check that the class we are looking for exists
    if class_name not in module_dict:
        valid_keys = list(module_dict.keys())
        raise ValueError(
            f"Could not find '{class_name}' in the dictionary "
            f"which maps names to classes. Available names: "
            f"{valid_keys}"
        )

    # Gather the arguments and keyword arguments to pass to the function
    args = config.pop("args", [])
    kwargs = dict(config.pop("kwargs", {}), **kwargs)

    # If args is specified as a dictionary, append it to kwargs (deprecated)
    if isinstance(args, dict):
        warn(
            "If specifying keyword arguments, should use `kwargs` instead "
            f"of `args` in {class_name}",
            category=DeprecationWarning,
        )
        for key in args.keys():
            assert key not in kwargs, (
                f"The keyword argument {key} is provided under "
                "`args` and `kwargs`. Ambiguous."
            )
        kwargs.update(args)
        args = []

    # If some arguments were specified at the top level, append them
    for key in config.keys():
        assert key not in kwargs, (
            f"The keyword argument {key} is provided "
            "at the top level and under `kwargs`. Ambiguous."
        )
    kwargs.update(config)

    # Intialize
    cls = module_dict[class_name]
    try:
        return cls(*args, **kwargs)

    except Exception as err:
        logger.error(
            f"Failed to instantiate {cls.__name__} with these arguments:\n"
            f"  - args: {args}\n  - kwargs: {kwargs}"
        )

        raise err



"""Module that defines an edge classification loss (ON vs OFF)."""

import numpy as np
import torch

from spine.model.layer.factories import loss_fn_factory
from spine.utils.enums import enum_factory
from spine.utils.globals import CLUST_COL, GROUP_COL, PART_COL, PRGRP_COL
from spine.utils.gnn.cluster import get_cluster_label_batch
from spine.utils.gnn.evaluation import (
    edge_assignment_batch,
    edge_assignment_from_graph_batch,
    edge_purity_mask_batch,
)
from spine.utils.weighting import get_class_weights

__all__ = ["EdgeChannelLoss"]


class EdgeChannelLoss(torch.nn.Module):
    """Takes the two-channel edge output of the GNN and optimizes edge-wise
    scores such that edges that connect nodes that belong to common instance
    are given a high score.

    For use in config:

    ..  code-block:: yaml

        model:
          name: grappa
          modules:
            grappa_loss:
              edge_loss:
                name: channel
                <dictionary of arguments to pass to the loss>

    See configuration files prefixed with `grappa_` under the `config`
    directory for detailed examples of working configurations.
    """

    # Name of the GNN loss (as specified in the configuration)
    name = "channel"

    def __init__(
        self, target, mode="group", loss="ce", balance_loss=False, high_purity=False
    ):
        """Initialize the primary identification loss function.

        Parameters
        ----------
        target : str
            Column name in the label tensor specifying the aggregation target
        mode : str, default 'group'
            Loss mode, one of 'group', 'forest' or 'particle_forest'
            - 'group' turns every edge that connect two nodes that belong to
              the same group (same target value) on
            - 'forest' ensures that at least one path in the graph connects two
              nodes, if they belong to the same group
            - 'particle_forest' only turns on edges that join two particles
              have a parentage relationship in the true particle tree
        loss : Union[str, dict], default 'ce'
            Name of the loss function to apply
        balance_loss : bool, default False
            Whether to weight the loss to account for class imbalance
        high_purity : bool, default False
            Only apply loss to nodes which belong to a sensible group, i.e.
            one with exactly one shower primary in it (not 0, not > 1)
        """
        # Initialize the parent class
        super().__init__()

        # Parse the aggregation target
        self.target = enum_factory("cluster", target)

        # Initialize basic parameters
        self.mode = mode
        self.balance_loss = balance_loss
        self.high_purity = high_purity

        # Set the loss
        self.loss_fn = loss_fn_factory(loss, functional=True)

    def forward(
        self, clust_label, clusts, edge_index, edge_pred, true_edge_index=None, **kwargs
    ):
        """Applies the edge channel loss to a batch of data.

        Parameters
        ----------
        clust_label : TensorBatch
            (N, 1 + D + N_f) Tensor of cluster labels for the batch
        clusts : IndexBatch
            (C) Index which maps each cluster to a list of voxel IDs
        edge_index : EdgeIndexBatch
            (2, E) Sparse ncidence matrix between clusters
        edge_pred : TensorBatch
            (E, 2) Edge prediction logits (binary output)
        true_edge_index : EdgeIndexBatch
            (2, E') True reference sparse incidence matrix
        **kwargs : dict, optional
            Other labels/outputs of the model which are not relevant here

        Returns
        -------
        loss : torch.Tensor
            Value of the loss
        accuracy : float
            Value of the edge-wise classification accuracy
        count : int
            Number of edges the loss was applied to
        """
        # Start to build a mask of valid edges. Check that the group ID
        # of both clusters edge joins has a valid group ID
        group_ids = get_cluster_label_batch(clust_label, clusts, self.target)
        valid_mask = np.all(group_ids.tensor[edge_index.index] > -1, axis=0)

        # If requested, check that each group contains a single shower primary
        if self.high_purity:
            assert self.target == GROUP_COL, (
                "The `high_purity` flag is only valid when " "building shower groups."
            )

            part_ids = get_cluster_label_batch(clust_label, clusts, PART_COL)
            prim_ids = get_cluster_label_batch(clust_label, clusts, PRGRP_COL)
            valid_mask &= edge_purity_mask_batch(
                edge_index, part_ids, group_ids, prim_ids
            )

        # Compute the edge assignments
        if self.mode == "group":
            # If an edge joins two nodes in the same group, label it as on
            edge_assn = edge_assignment_batch(edge_index, group_ids)

        elif self.mode == "forest":
            # For each group, find the most likely spanning tree, label the
            # edges in the tree as 1. For all other edges, apply loss only if
            # in separate groups. If undirected, also assign symmetric path
            edge_assn, valid_mask_mst = edge_assignment_forest_batch(
                edge_index, edge_pred.to_numpy(), group_ids
            )
            valid_mask &= valid_mask_mst

        elif self.mode == "particle_forest":
            # If an edge matches a parentage relation, mark it as on
            assert true_edge_index is not None, (
                "Must provide true `true_edge_index` object when using "
                "the `particle_forest` truth mode"
            )

            part_ids = get_cluster_label_batch(clust_label, clusts, PART_COL)
            edge_assn = edge_assignement_from_graph_batch(
                edge_index, true_edge_index, part_ids
            )

        else:
            raise ValueError("Loss mode not recognized:", self.mode)

        # Apply the mask and convert the labels to a torch.Tensor
        valid_index = np.where(valid_mask)[0]
        edge_assn = edge_assn.to_tensor(dtype=torch.long, device=edge_pred.device)
        edge_pred = edge_pred.tensor[valid_index]
        edge_assn = edge_assn.tensor[valid_index]

        # Compute the loss. Balance classes if requested
        weights = None
        if self.balance_loss:
            weights = get_class_weights(edge_assn, num_classes=2)

        loss = self.loss_fn(edge_pred, edge_assn, weight=weights, reduction="sum")
        if len(valid_index):
            loss /= len(valid_index)

        # Compute accuracy of assignment (fraction of correctly assigned edges)
        acc = 1.0
        if len(valid_index):
            acc = float(torch.sum(torch.argmax(edge_pred, dim=1) == edge_assn))
            acc /= len(valid_index)

        return {"accuracy": acc, "loss": loss, "count": len(valid_index)}



"""Module which supports likelihood-based flash matchin (OpT0Finder)."""

import os
import sys

import numexpr as ne
import numpy as np


class LikelihoodFlashMatcher:
    """Interface class between full chain outputs and OpT0Finder

    See https://github.com/drinkingkazu/OpT0Finder for more details about it.
    """

    def __init__(
        self,
        cfg,
        detector,
        parent_path=None,
        scaling=1.0,
        alpha=0.21,
        recombination_mip=0.65,
        legacy=False,
    ):
        """Initialize the likelihood-based flash matching algorithm.

        Parameters
        ----------
        cfg : str
            Flash matching configuration file path
        detector : str, optional
            Detector to get the optical detector properties for
        parent_path : str, optional
            Path to the parent configuration file (allows for relative paths)
        scaling : Union[float, str], default 1.
            Global scaling factor for the depositions (can be an expression)
        alpha : float, default 0.21
            Number of excitons (Ar*) divided by number of electron-ion pairs (e-,Ar+)
        recombination_mip : float, default 0.65
            Recombination factor for MIP-like particles in LAr
        legacy : bool, default False
            Use the legacy OpT0Finder function(s). TODO: remove when dropping legacy
        """
        # Initialize the flash manager (OpT0Finder wrapper)
        self.initialize_backend(cfg, detector, parent_path)

        # Get the external parameters
        self.scaling = scaling
        if isinstance(self.scaling, str):
            self.scaling = ne.evaluate(self.scaling)
        self.alpha = alpha
        if isinstance(self.alpha, str):
            self.alpha = ne.evaluate(self.alpha)
        self.recombination_mip = recombination_mip
        if isinstance(self.recombination_mip, str):
            self.recombination_mip = ne.evaluate(self.recombination_mip)
        self.legacy = legacy

        # Initialize flash matching attributes
        self.matches = None
        self.qcluster_v = None
        self.flash_v = None

    def initialize_backend(self, cfg, detector, parent_path):
        """Initialize OpT0Finder (backend).

        Expects that the environment variable `FMATCH_BASEDIR` is set.
        You can either set it by hand (to the path where one can find
        OpT0Finder) or you can source `OpT0Finder/configure.sh` if you
        are running code from a command line.

        Parameters
        ----------
        cfg: str
            Path to config for OpT0Finder
        detector : str, optional
            Detector to get the optical detector properties for
        parent_path : str, optional
            Path to the parent configuration file (allows for relative paths)
        """
        # Add OpT0finder python interface to the python path
        basedir = os.getenv("FMATCH_BASEDIR")
        assert basedir is not None, (
            "You need to source OpT0Finder's configure.sh or set the "
            "FMATCH_BASEDIR environment variable before running flash "
            "matching."
        )
        sys.path.append(os.path.join(basedir, "python"))

        # Add the OpT0Finder library to the dynamic link loader
        lib_path = os.path.join(basedir, "build/lib")
        os.environ["LD_LIBRARY_PATH"] = "{}:{}".format(
            lib_path, os.environ["LD_LIBRARY_PATH"]
        )

        # Add the OpT0Finder data directory if it is not yet set
        if "FMATCH_DATADIR" not in os.environ:
            os.environ["FMATCH_DATADIR"] = os.path.join(basedir, "dat")

        # Load up the detector specifications
        if detector is None:
            det_cfg = os.path.join(basedir, "dat/detector_specs.cfg")
        else:
            det_cfg = os.path.join(basedir, f"dat/detector_specs_{detector}.cfg")

        if not os.path.isfile(det_cfg):
            raise FileNotFoundError(
                f"Cannot file detector specification file: {det_cfg}."
            )

        from flashmatch import flashmatch

        flashmatch.DetectorSpecs.GetME(det_cfg)

        # Fetch and initialize the OpT0Finder configuration
        if parent_path is not None and not os.path.isfile(cfg):
            cfg = os.path.join(parent_path, cfg)
        if not os.path.isfile(cfg):
            raise FileNotFoundError(f"Cannot find flash-matcher config: {cfg}")

        cfg = flashmatch.CreateFMParamsFromFile(cfg)

        # Initialize The OpT0Finder flash match manager
        self.mgr = flashmatch.FlashMatchManager()
        self.mgr.Configure(cfg)

        # Get the light path algorithm to produce QCluster_t objects
        self.light_path = flashmatch.CustomAlgoFactory.get().create(
            "LightPath", "ToyMCLightPath"
        )
        self.light_path.Configure(cfg.get["flashmatch::FMParams"]("LightPath"))

    def get_matches(self, interactions, flashes):
        """Find TPC interactions compatible with optical flashes.

        Parameters
        ----------
        interactions : List[Union[Interaction, TruthInteraction]]
            List of TPC interactions
        flashes : List[Flash]
            List of optical flashes

        Returns
        -------
        List[Tuple[Interaction, Flash, flashmatch::FlashMatch_t]]
            Set of interaction/flash matches with their matching characteristics
        """
        # If there is no interaction or no flashe, nothing to do
        if not len(interactions) or not len(flashes):
            return []

        # Build a list of QCluster_t (OpT0Finder interaction representation)
        self.qcluster_v = self.make_qcluster_list(interactions)

        # Build a list of Flash_t (OpT0Finder optical flash representation)
        self.flash_v, flashes = self.make_flash_list(flashes)

        # Running flash matching and caching the results
        self.matches = self.run_flash_matching()

        # Build result, return
        result = []
        for m in self.matches:
            tpc_id = self.qcluster_v[m.tpc_id].idx
            flash_id = self.flash_v[m.flash_id].idx
            result.append((interactions[tpc_id], flashes[flash_id], m))

        return result

    def make_qcluster_list(self, interactions):
        """Converts a list of SPINE interaction into a list of OpT0Finder
        flashmatch.QCluster_t objects.

        Parameters
        ----------
        interactions : List[Union[Interaction, TruthInteraction]]
            List of TPC interactions

        Returns
        -------
        List[QCluster_t]
           List of OpT0Finder flashmatch::QCluster_t objects
        """
        # Loop over the interacions
        from flashmatch import flashmatch

        qcluster_v = []
        for idx, inter in enumerate(interactions):
            # Produce a mask to remove negative value points (can happen)
            valid_mask = np.where(inter.depositions > 0.0)[0]

            # Skip interactions with less than 2 points
            if len(valid_mask) < 2:
                continue

            # Initialize qcluster
            qcluster = flashmatch.QCluster_t()
            qcluster.idx = idx
            qcluster.time = 0

            # Get the point coordinates
            points = inter.points[valid_mask]

            # Get the depositions
            depositions = inter.depositions[valid_mask]

            # Fill the trajectory
            pytraj = np.hstack([points, depositions[:, None]])
            traj = flashmatch.as_geoalgo_trajectory(pytraj)
            if self.legacy:
                qcluster += self.light_path.MakeQCluster(traj, self.scaling)
            else:
                qcluster += self.light_path.MakeQCluster(
                    traj, self.scaling, self.alpha, self.recombination_mip
                )

            # Append
            qcluster_v.append(qcluster)

        return qcluster_v

    def make_flash_list(self, flashes):
        """Creates a list of flashmatch.Flash_t from the local class.

        Parameters
        ----------
        flashes : List[Flash]
            List of optical flashes

        Returns
        -------
        List[Flash_t]
            List of flashmatch::Flash_t objects
        """
        # Loop over the optical flashes
        from flashmatch import flashmatch

        flash_v = []
        for idx, f in enumerate(flashes):
            # Initialize the Flash_t object
            flash = flashmatch.Flash_t()
            flash.idx = idx
            flash.time = f.time

            # Assign the flash position and error on this position
            flash.x, flash.y, flash.z = 0, 0, 0
            flash.x_err, flash.y_err, flash.z_err = 0, 0, 0

            # Assign the individual PMT optical hit PEs
            for pe in f.pe_per_ch:
                flash.pe_v.push_back(pe)
                flash.pe_err_v.push_back(0.0)

            # Append
            flash_v.append(flash)

        return flash_v, flashes

    def run_flash_matching(self):
        """Drive the OpT0Finder flash matching.

        Returns
        -------
        List[flashmatch::FlashMatch_t]
            List of matches
        """
        # Make sure the interaction and flash objects were set
        assert (
            self.qcluster_v is not None and self.flash_v is not None
        ), "Must make_qcluster_list and make_flash_list first."

        # Register all objects in the manager
        self.mgr.Reset()
        for x in self.qcluster_v:
            self.mgr.Add(x)
        for x in self.flash_v:
            self.mgr.Add(x)

        # Run the matching
        all_matches = self.mgr.Match()

        # Adjust the output position to account for the module shift
        for m in all_matches:
            pos = np.array([m.tpc_point.x, m.tpc_point.y, m.tpc_point.z])
            m.tpc_point.x = pos[0]
            m.tpc_point.y = pos[1]
            m.tpc_point.z = pos[2]

        return all_matches

    def get_qcluster(self, idx, array=False):
        """Fetch a given flashmatch::QCluster_t object.

        Parameters
        ----------
        idx : int
            ID of the interaction to fetch
        array : bool, default `False`
            If `True`, The QCluster is returned as an np.ndarray

        Returns
        -------
        Union[flashmatch::QCluster_t, np.ndarray]
            QCluster object
        """
        from flashmatch import flashmatch

        if self.qcluster_v is None:
            raise ValueError("self.qcluster_v is None")

        for qcluster in self.qcluster_v:
            if qcluster.idx != idx:
                continue
            if array:
                return flashmatch.as_ndarray(qcluster)
            else:
                return qcluster

        raise IndexError(f"TPC object {idx} does not exist in self.qcluster_v")

    def get_flash(self, idx, array=False):
        """Fetch a given flashmatch::Flash object.

        Parameters
        ----------
        idx : int
            ID of the flash to fetch
        array : bool, default `False`
            If `True`, The flash is returned as an np.ndarray

        Returns
        -------
        Union[flashmatch::Flash, np.ndarray]
            Flash object
        """
        from flashmatch import flashmatch

        if self.flash_v is None:
            raise ValueError("self.flash_v is None")

        for flash in self.flash_v:
            if flash.idx != idx:
                continue
            if array:
                return flashmatch.as_ndarray(flash)
            else:
                return flash

        raise IndexError(f"Flash {idx} does not exist in self.flash_v")

    def get_match(self, idx):
        """Fetch a match for a given TPC interaction ID.

        Parameters
        ----------
        idx : int
            Index of TPC object for which we want to retrieve a match

        Returns
        -------
        flashmatch::FlashMatch_t
            Flash match associated with interaction idx
        """
        if self.matches is None or self.qcluster_v is None:
            raise ValueError("Need to run flash matching first")

        for m in self.matches:
            if self.qcluster_v[m.tpc_id].idx != idx:
                continue
            return m

        return None

    def get_matched_flash(self, idx):
        """Fetch a matched flash for a given TPC interaction ID.

        Parameters
        ----------
        idx : int
            Index of TPC object for which we want to retrieve a match

        Returns
        -------
        flashmatch::Flash_t
            Optical flash that matches interaction idx
        """
        # Check if we have already run flash matching
        if self.flash_v is None or self.matches is None:
            raise ValueError("Need to run flash matching first")

        # Get a match, if any
        m = self.get_match(idx)
        if m is None:
            return None

        # Get the flash that corresponds to the match
        flash_id = m.flash_id
        if flash_id is None:
            return None
        if flash_id >= len(self.flash_v):
            raise IndexError(f"Flash {flash_id} does not exist in self.flash_v")

        return self.flash_v[flash_id]

    def get_t0(self, idx):
        """Fetch a matched flash time for a given TPC interaction ID.

        Parameters
        ----------
        idx : int
            Index of TPC object for which we want to retrieve a match

        Returns
        -------
        float
            Time in us with respect to simulation time reference
        """
        # Get the matched flash, if any
        flash = self.get_matched_flash(idx)

        return None if flash is None else flash.time



"""Tools to draw a (labeled) point cloud."""

import numpy as np
import plotly.graph_objs as go

from spine.utils.globals import COORD_COLS

__all__ = ["scatter_points"]


def scatter_points(
    points,
    color=None,
    markersize=2,
    linewidth=2,
    colorscale=None,
    cmin=None,
    cmax=None,
    opacity=None,
    hovertext=None,
    hovertemplate=None,
    dim=3,
    mode="markers",
    marker=None,
    line=None,
    **kwargs,
):
    """Scatters points and their labels.

    Produces :class:`plotly.graph_objs.Scatter3d` or
    :class:`plotly.graph_objs.Scatter` trace object to be drawn in plotly. The
    object is nested to be fed directly to a :class:`plotly.graph_objs.Figure`
    or :func:`plotly.offline.iplot`. All of the regular plotly parameters are
    available.

    It can scatter points individually (default `mode`) or it can draw lines
    between the provided points (`mode='lines'` option).

    Parameters
    ----------
    points : np.ndarray
        (N, 2+) array of N points of (..., x, y, [z],...) coordinate information
    color : Union[str, np.ndarray], optional
        Color of markers/lines or (N) list of color of markers/lines
    markersize : float, default 2
        Marker size
    linewidth : float, default 2
        Line width
    colorscale : Union[str, List[str], List[List[float, str]], optional
        Plotly colorscale specifier for the markers
    cmin : Union[int, float], optional
        Minimum of the color range
    cmax : Union[int, float], optional
        Maximum of the color range
    opacity : float
        Marker opacity
    hovertext : Union[List[str], List[int]], optional
        (N) List of labels associated with each marker
    hovertemplate : str, optional
        Hover information formatting
    dim : int, default 3
        Dimension (can either be 2 or 3)
    mode : str, default 'markers'
        Drawing mode
    marker : dict, optional
        Marker style configuration dictionary
    line : dict, optional
       Line style configuration dictionary
    **kwargs : dict, optional
        List of additional arguments to pass to plotly.graph_objs.Scatter3D

    Returns
    -------
    List[go.Scatter3d]
        (1) List with one graph of the input points
    """
    # Check the dimension for compatibility
    if dim not in [2, 3]:
        raise ValueError("This function only supports dimension 2 or 3.")
    if points.shape[1] == 2:
        dim = 2

    # Get the coordinate column locations in the input tensor
    coord_cols = COORD_COLS
    if dim == 2:
        coord_cols = COORD_COLS[:2]
    if points.shape[1] == dim:
        coord_cols = np.arange(dim)

    # If there is no hovertext, print the color as part of the hovertext
    if hovertext is None and color is not None and not isinstance(color, str):
        hovertext = [f"Value: {c}" for c in color]

    # Update hovertemplate
    if hovertemplate is None:
        hovertemplate = "x: %{x}<br>y: %{y}"
        if dim == 3:
            hovertemplate += "<br>z: %{z}"
        if hovertext is not None:
            if not np.isscalar(hovertext):
                hovertemplate += "<br>%{text}"
            else:
                hovertemplate += f"<br>{hovertext}"
                hovertext = None

    # Initialize the marker/line object depending on mode
    if "markers" in mode and marker is None:
        marker = {
            "size": markersize,
            "color": color,
            "opacity": opacity,
            "colorscale": colorscale,
            "cmin": cmin,
            "cmax": cmax,
        }
    if "lines" in mode and line is None:
        if dim == 2:
            line = {"width": linewidth, "color": color}
        else:
            line = {
                "width": linewidth,
                "color": color,
                "colorscale": colorscale,
                "cmin": cmin,
                "cmax": cmax,
            }

    # Initialize and return
    axes = ["x", "y", "z"][:dim]
    pos_dict = {a: points[:, coord_cols[i]] for i, a in enumerate(axes)}

    if dim == 2:
        return [
            go.Scatter(
                mode=mode,
                marker=marker,
                line=line,
                text=hovertext,
                hovertemplate=hovertemplate,
                **pos_dict,
                **kwargs,
            )
        ]
    else:
        return [
            go.Scatter3d(
                mode=mode,
                marker=marker,
                line=line,
                text=hovertext,
                hovertemplate=hovertemplate,
                **pos_dict,
                **kwargs,
            )
        ]



"""Reconstruction post-processor modules."""

from .calo import *
from .cathode_cross import *
from .cluster import *
from .direction import *
from .geometry import *
from .kinematics import *
from .mcs import *
from .pid import *
from .points import *
from .ppn import *
from .shower import *
from .source import *
from .topology import *
from .tracking import *
from .vertex import *



"""Contains a reader class dedicated to loading data from HDF5 files."""

from dataclasses import fields
from typing import Any, Dict, List, Optional, Union
from warnings import warn

import h5py
import numpy as np
import yaml
from yaml.parser import ParserError

import spine.data
from spine.utils.docstring import inherit_docstring
from spine.utils.logger import logger

from .base import ReaderBase

__all__ = ["HDF5Reader"]


@inherit_docstring(ReaderBase)
class HDF5Reader(ReaderBase):
    """Class which reads information stored in HDF5 files.

    This class inherits from the :class:`ReaderBase` class. It provides
    methods to load HDF5 files and extract their data products. The files
    must be structured as follows:
      - An `events` dataset with all the region references
      - One dataset per data product corresponding to each region reference in
        the `events` dataset
    """

    name = "hdf5"

    def __init__(
        self,
        file_keys: Optional[Union[str, List[str]]] = None,
        file_list: Optional[str] = None,
        limit_num_files: Optional[int] = None,
        max_print_files: int = 10,
        n_entry: Optional[int] = None,
        n_skip: Optional[int] = None,
        entry_list: Optional[List[int]] = None,
        skip_entry_list: Optional[List[int]] = None,
        run_event_list: Optional[List[List[int]]] = None,
        skip_run_event_list: Optional[List[List[int]]] = None,
        create_run_map: bool = False,
        build_classes: bool = True,
        skip_unknown_attrs: bool = False,
        run_info_key: str = "run_info",
        allow_missing: bool = False,
    ) -> None:
        """Initalize the HDF5 file reader.

        Parameters
        ----------
        file_keys : Union[str, List[str]], optional
            Path or list of paths to the HDF5 files to be read
        file_list : str, optional
            Path to a text file containing a list of file paths to be read
        limit_num_files : Optional[int], optional
            Integer limiting number of files to be taken per data directory
        max_print_files : int, default 10
            Maximum number of loaded file names to be printed
        n_entry : Optional[int], optional
            Maximum number of entries to load
        n_skip : Optional[int], optional
            Number of entries to skip at the beginning
        entry_list : Optional[List[int]]
            List of integer entry IDs to add to the index
        skip_entry_list : Optional[List[int]]
            List of integer entry IDs to skip from the index
        run_event_list: Optional[List[List[int]]], optional
            List of (run, subrun, event) triplets to add to the index
        skip_run_event_list: Optional[List[List[int]]], optional
            List of (run, subrun, event) triplets to skip from the index
        create_run_map : bool, default False
            Initialize a map between (run, subrun, event) triplets and entries.
            For large files, this can be quite expensive (must load every entry).
        build_classes : bool, default True
            If the stored object is a class, build it back
        skip_unknown_attrs : bool, default False
            If `True`, allow a loaded object to have unrecognized attributes.
            This allows backward compatibility with old files, but use with
            extreme caution, as this might hide a fundamental issue with your code.
        run_info_key : str, default 'run_info'
            Name of the data product which contains the run info of the event
        allow_missing : bool, default False
            If `True`, allows missing entries in the entry or event list
        """
        # Process the list of files
        self.process_file_paths(file_keys, file_list, limit_num_files, max_print_files)

        # If an entry list is requested based on run/subrun/event ID, create map
        if run_event_list is not None or skip_run_event_list is not None:
            create_run_map = True

        # Loop over the input files, build a map from index to file ID
        file_index, run_info = [], []
        self.num_entries = 0
        self.file_offsets = np.empty(len(self.file_paths), dtype=np.int64)
        for i, path in enumerate(self.file_paths):
            with h5py.File(path, "r") as in_file:
                # Check that there are events in the file
                assert "events" in in_file, "File does not contain an event tree"

                events = in_file["events"]
                assert isinstance(
                    events, h5py.Dataset
                ), "'events' is not a dataset in the HDF5 file."

                # If requested, register the (run, subrun, event) information
                if create_run_map:
                    assert (
                        run_info_key in in_file
                    ), f"Must provide {run_info_key} to create run map"

                    info = in_file[run_info_key]
                    assert isinstance(
                        info, h5py.Dataset
                    ), f"{run_info_key} is not a dataset in the HDF5 file."
                    assert all(
                        k in info.dtype.names for k in ["run", "subrun", "event"]
                    ), f"{run_info_key} dataset missing required fields."

                    for r, s, e in zip(info["run"], info["subrun"], info["event"]):
                        run_info.append((r, s, e))

                # Update the total number of entries
                num_entries = len(events)
                file_index.append(i * np.ones(num_entries, dtype=np.int64))
                self.file_offsets[i] = self.num_entries
                self.num_entries += num_entries

        # Dump the number of entries to load
        logger.info("Total number of entries in the file(s): %d\n", self.num_entries)

        # Concatenate the file indexes into one, set run info if needed
        self.file_index = np.concatenate(file_index)
        self.run_info = run_info if create_run_map else None

        # Process the run information
        self.process_run_info()

        # Process the entry list
        self.process_entry_list(
            n_entry,
            n_skip,
            entry_list,
            skip_entry_list,
            run_event_list,
            skip_run_event_list,
            allow_missing,
        )

        # Store other attributes
        self.build_classes = build_classes
        self.skip_unknown_attrs = skip_unknown_attrs

        # Process the configuration used to produce the HDF5 file
        self.cfg = self.process_cfg()

        # Process the SPINE version used to produced the HDF5 file
        self.version = self.process_version()

    def process_cfg(self) -> Optional[dict]:
        """Fetches the SPINE configuration used to produce the HDF5 file.

        Returns
        -------
        dict
            Configuration dictionary
        """
        # Fetch the string-form configuration
        with h5py.File(self.file_paths[0], "r") as in_file:
            assert "info" in in_file, "HDF5 file missing 'info' group."
            assert (
                "cfg" in in_file["info"].attrs
            ), "HDF5 file 'info' group missing 'cfg' attribute."
            cfg_str = in_file["info"].attrs["cfg"]

        # Attempt to parse it (need try for now for SPINE versions < v0.4.0)
        try:
            assert isinstance(cfg_str, str), "'cfg' attribute is not a string."
            cfg = yaml.safe_load(cfg_str)
        except ParserError:
            warn(
                "Parsing configuration failed, returning None for SPINE versions < v0.4.0"
            )
            return None

        return cfg

    def process_version(self) -> str:
        """Returns the SPINE release version used to produce the HDF5 file.

        Returns
        -------
        str
            SPINE release tag
        """
        # Fetch the string-form configuration
        with h5py.File(self.file_paths[0], "r") as in_file:
            assert "info" in in_file, "HDF5 file missing 'info' group."
            assert (
                "version" in in_file["info"].attrs
            ), "HDF5 file 'info' group missing 'version' attribute."
            version = in_file["info"].attrs["version"]

        assert isinstance(version, str), "'version' attribute is not a string."
        return version

    def get(self, idx: int) -> Dict[str, Any]:
        """Returns a specific entry in the file.

        Parameters
        ----------
        idx : int
            Integer entry ID to access

        Returns
        -------
        data : dict
            Ditionary of data products corresponding to one event
        """
        # Get the appropriate entry index
        assert idx < len(self.entry_index)
        file_idx = self.get_file_index(idx)
        entry_idx = self.get_file_entry_index(idx)

        # Use the event tree to find out what needs to be loaded
        data = {"file_index": file_idx, "file_entry_index": entry_idx}
        with h5py.File(self.file_paths[file_idx], "r") as in_file:
            events = in_file["events"]
            assert isinstance(
                events, h5py.Dataset
            ), "'events' is not a dataset in the HDF5 file."

            event = events[entry_idx]
            names = getattr(getattr(event, "dtype", None), "names", None)
            if names is not None:
                for key in names:
                    self.load_key(in_file, event, data, key)
            else:
                raise ValueError("Event entry does not have named fields.")

        # Use the global index, not the one read from file
        data["index"] = idx

        return data

    def load_key(
        self, in_file: h5py.File, event: Dict[str, Any], data: Dict[str, Any], key: str
    ) -> None:
        """Fetch a specific key for a specific event.

        Parameters
        ----------
        in_file : h5py.File
            HDF5 file instance
        event : dict
            Dictionary of objects that make up one event
        data : dict
            Dictionary of data products corresponding to one event
        key: str
            Name of the dataset in the entry
        """
        # The event-level information is a region reference: fetch it
        region_ref = event[key]
        dataset = in_file[key]
        if isinstance(dataset, h5py.Dataset):
            names = getattr(getattr(dataset, "dtype", None), "names", None)
            if not names:
                # If the reference points at a simple dataset, return
                data[key] = dataset[region_ref]
                if dataset.attrs["scalar"]:
                    data[key] = data[key][0]
                if len(dataset.shape) > 1:
                    data[key] = data[key].reshape(-1, dataset.shape[1])

            else:
                # If the dataset has multiple attributes, it contains an object.
                # Start by fetching the appropriate class to rebuild
                array = dataset[region_ref]
                class_name = dataset.attrs["class_name"]
                assert isinstance(
                    class_name, str
                ), "Dataset missing 'class_name' attribute."
                obj_class = getattr(spine.data, class_name)

                # If needed, get the list of recognized attributes
                known_attrs = []
                if self.skip_unknown_attrs:
                    known_attrs = [f.name for f in fields(obj_class)]

                # Load the object
                names = array.dtype.names
                data[key] = []
                for i, el in enumerate(array):
                    # Fetch the list of key/value pairs, filter if requested
                    if self.skip_unknown_attrs:
                        obj_dict = {}
                        for i, k in enumerate(names):
                            if k in known_attrs:
                                obj_dict[k] = el[i]
                    else:
                        obj_dict = dict(zip(names, el))

                    # Rebuild an instance of the object class, if requested
                    if self.build_classes:
                        data[key].append(obj_class(**obj_dict))
                    else:
                        data[key].append(obj_dict)

                if in_file[key].attrs["scalar"]:
                    data[key] = data[key][0]

        elif isinstance(dataset, h5py.Group):
            # If the reference points at a group, unpack
            index = dataset["index"]
            assert isinstance(index, h5py.Dataset), "Dataset 'index' is missing."
            el_refs = index[region_ref].flatten()
            if len(index.shape) == 1:
                elements = dataset["elements"]
                assert isinstance(
                    elements, h5py.Dataset
                ), "Dataset 'elements' is missing."
                ret = np.empty(len(el_refs), dtype=object)
                ret[:] = [elements[r] for r in el_refs]
                if len(elements.shape) > 1:
                    for i in range(len(el_refs)):
                        ret[i] = ret[i].reshape(-1, elements.shape[1])

            else:
                elements = [dataset[f"element_{i}"] for i in range(len(el_refs))]
                ret = []
                for i, el in enumerate(elements):
                    assert isinstance(
                        el, h5py.Dataset
                    ), f"Dataset 'element_{i}' is missing."
                    ret.append(el[el_refs[i]])
                    if len(el.shape) > 1:
                        ret[i] = ret[i].reshape(-1, el.shape[1])

            data[key] = ret

        else:
            raise ValueError(f"Dataset for key '{key}' is neither a group nor dataset.")



"""Base class of all analysis scripts."""

from abc import ABC, abstractmethod
from warnings import warn

from spine.io.core.write.csv import CSVWriter


class AnaBase(ABC):
    """Parent class of all analysis scripts.

    This base class performs the following functions:
    - Ensures that the necessary methods exist
    - Checks that the script is provided the necessary information
    - Writes the output of the analysis to CSV

    Attributes
    ----------
    name : str
        Name of the analysis script (to call it from a configuration file)
    req_keys : List[str]
        Data products needed to run the analysis script
    opt_keys : List[str]
        Optional data products which will be used only if they are provided
    units : str
        Units in which the coordinates are expressed
    """

    # Name of the analysis script (as specified in the configuration)
    name = None

    # Alternative allowed names of the analysis script
    aliases = ()

    # Units in which the analysis script expects objects to be expressed in
    units = "cm"

    # Set of data keys needed for this analysis script to operate
    _keys = ()

    # List of recognized object types
    _obj_types = ("fragment", "particle", "interaction")

    # Valid run modes
    _run_modes = ("reco", "truth", "both", "all")

    # List of known point modes for true particles and their corresponding keys
    _point_modes = (
        ("points", "points_label"),
        ("points_adapt", "points"),
        ("points_g4", "points_g4"),
    )

    # List of known deposition modes for true particles and their corresponding keys
    _dep_modes = (
        ("depositions", "depositions_label"),
        ("depositions_q", "depositions_q_label"),
        ("depositions_adapt", "depositions_label_adapt"),
        ("depositions_adapt_q", "depositions"),
        ("depositions_g4", "depositions_g4"),
    )

    def __init__(
        self,
        obj_type=None,
        run_mode=None,
        truth_point_mode=None,
        truth_dep_mode=None,
        append=False,
        overwrite=False,
        log_dir=None,
        prefix=None,
    ):
        """Initialize default anlysis script object properties.

        Parameters
        ----------
        obj_type : Union[str, List[str]]
            Name or list of names of the object types to process
        run_mode : str, optional
            If specified, tells whether the analysis script must run on
            reconstructed ('reco'), true ('true') or both objects
            ('both' or 'all')
        truth_point_mode : str, optional
            If specified, tells which attribute of the :class:`TruthFragment`,
            :class:`TruthParticle` or :class:`TruthInteraction` object to use
            to fetch its point coordinates
        truth_dep_mode : str, optional
            If specified, tells which attribute of the :class:`TruthFragment`,
            :class:`TruthParticle` or :class:`TruthInteraction` object to use
            to fetch its energy depositions
        append : bool, default False
            If True, appends existing CSV files instead of creating new ones
        overwrite : bool, default False
            If True and an output CSV file exists, overwrite it
        log_dir : str
            Output CSV file directory (shared with driver log)
        prefix : str, default None
            Name to prefix every output CSV file with
        """
        # Initialize default keys
        self.update_keys(
            {
                "index": True,
                "file_index": True,
                "file_entry_index": False,
                "run_info": False,
            }
        )

        # If run mode is specified, process it
        self.run_mode = run_mode
        if run_mode is not None:
            # Check that the run mode is recognized
            assert run_mode in self._run_modes, (
                f"`run_mode` not recognized: {run_mode}. Must be one of "
                f"{self._run_modes}."
            )

        self.prefixes = []
        if run_mode != "truth":
            self.prefixes.append("reco")
        if run_mode != "reco":
            self.prefixes.append("truth")

        # Check that all the object sources are recognized
        self.obj_type = obj_type
        if self.obj_type is not None:
            if isinstance(self.obj_type, str):
                self.obj_type = [self.obj_type]
            for obj in self.obj_type:
                assert obj in self._obj_types, (
                    f"Object type must be one of {self._obj_types}. Got "
                    f"`{obj}` instead."
                )

        # Make a list of object keys to process
        self.fragment_keys = []
        self.particle_keys = []
        self.interaction_keys = []
        for name in self._obj_types:
            # Initialize one list per object type
            setattr(self, f"{name}_keys", [])

            # Skip object types which are not requested
            if self.obj_type is not None and name in self.obj_type:
                if run_mode != "truth":
                    getattr(self, f"{name}_keys").append(f"reco_{name}s")
                if run_mode != "reco":
                    getattr(self, f"{name}_keys").append(f"truth_{name}s")

        self.obj_keys = self.fragment_keys + self.particle_keys + self.interaction_keys

        # Update underlying keys, if needed
        self.update_keys({k: True for k in self.obj_keys})

        # If a truth point mode is specified, store it
        if truth_point_mode is not None:
            assert truth_point_mode in self.point_modes, (
                "The `truth_point_mode` argument must be one of "
                f"{self.point_modes.keys()}. Got `{truth_point_mode}` instead."
            )
            self.truth_point_mode = truth_point_mode
            self.truth_index_mode = truth_point_mode.replace("points", "index")

        # If a truth deposition mode is specified, store it
        if truth_dep_mode is not None:
            assert truth_dep_mode in self.dep_modes, (
                "The `truth_dep_mode` argument must be one of "
                f"{self.dep_modes.keys()}. Got `{truth_dep_mode}` instead."
            )
            if truth_point_mode is not None:
                prefix = truth_point_mode.replace("points", "depositions")
                assert truth_dep_mode.startswith(prefix), (
                    f"Points mode {truth_point_mode} and deposition mode "
                    f"{truth_dep_mode} are incompatible."
                )
            self.truth_dep_mode = truth_dep_mode
            self.truth_dep_key = self.dep_modes[truth_dep_mode]

        # Store the append flag
        self.append_file = append
        self.overwrite_file = overwrite

        # Initialize a writer dictionary to be filled by the children classes
        self.base_dict = {}
        self.log_dir = log_dir
        self.output_prefix = prefix
        self.writers = {}

    def initialize_writer(self, name):
        """Adds a CSV writer to the list of writers for this script.

        Parameters
        ----------
        name : str
            Name of the writer
        """
        # Define the name of the file to write to
        assert len(name) > 0, "Must provide a non-empty name."
        file_name = f"{self.name}_{name}.csv"
        if self.output_prefix:
            file_name = f"{self.output_prefix}_{file_name}"
        if self.log_dir:
            file_name = f"{self.log_dir}/{file_name}"

        # Initialize the writer
        self.writers[name] = CSVWriter(
            file_name, append=self.append_file, overwrite=self.overwrite_file
        )

    @property
    def keys(self):
        """Dictionary of (key, necessity) pairs which determine which data keys
        are needed/optional for the post-processor to run.

        Returns
        -------
        Dict[str, bool]
            Dictionary of (key, necessity) pairs to be used
        """
        return dict(self._keys)

    @keys.setter
    def keys(self, keys):
        """Converts a dictionary of keys to an immutable tuple.

        Parameters
        ----------
        Dict[str, bool]
            Dictionary of (key, necessity) pairs to be used
        """
        self._keys = tuple(keys.items())

    @property
    def point_modes(self):
        """Dictionary which makes the correspondance between the name of a true
        object point attribute with the underlying point tensor it points to.

        Returns
        -------
        Dict[str, str]
            Dictionary of (attribute, key) mapping for point coordinates
        """
        return dict(self._point_modes)

    @property
    def dep_modes(self):
        """Dictionary which makes the correspondance between the name of a true
        object deposition attribute with the underlying deposition array it points to.

        Returns
        -------
        Dict[str, str]
            Dictionary of (attribute, key) mapping for point depositions
        """
        return dict(self._dep_modes)

    def update_keys(self, update_dict):
        """Update the underlying set of keys and their necessity in place.

        Parameters
        ----------
        update_dict : Dict[str, bool]
            Dictionary of (key, necessity) pairs to update the keys with
        """
        if len(update_dict) > 0:
            keys = self.keys
            keys.update(update_dict)
            self._keys = tuple(keys.items())

    def get_base_dict(self, data):
        """Builds the entry information dictionary.

        Parameters
        ----------
        data : dict
            Dictionary of data products

        Returns
        -------
        dict
            Dictionary of information for this entry
        """
        # Extract basic information to store in every row
        base_dict = {"index": data["index"], "file_index": data["file_index"]}
        if "file_entry_index" in data:
            base_dict["file_entry_index"] = data["file_entry_index"]
        if "run_info" in data:
            base_dict.update(**data["run_info"].scalar_dict())
        else:
            warn("`run_info` is missing; will not be included in CSV file.")

        return base_dict

    def append(self, name, **kwargs):
        """Apppend a CSV log file with a set of values.

        Parameters
        ----------
        name : str
            Name of the writer
        **kwargs : dict
            Dictionary of information to save to the writer
        """
        self.writers[name].append({**self.base_dict, **kwargs})

    def __call__(self, data, entry=None):
        """Runs the analysis script on one entry.

        Parameters
        ----------
        data : dict
            Data dictionary for one entry

        Returns
        -------
        dict
            Update to the input dictionary
        """
        # Fetch the necessary information
        data_filter = {}
        for key, req in self.keys.items():
            # If this key is needed, check that it exists
            assert not req or key in data, (
                f"Analysis script `{self.name}` is missing an essential "
                f"input to be used: `{key}`."
            )

            # Append
            if key in data:
                data_filter[key] = data[key]
                if entry is not None:
                    data_filter[key] = data[key][entry]

        # Fetch the base dictionary
        self.base_dict = self.get_base_dict(data_filter)

        # Run the analysis script
        return self.process(data_filter)

    def get_index(self, obj):
        """Get a certain pre-defined index attribute of an object.

        The :class:`TruthFragment`, :class:`TruthParticle` and
        :class:`TruthInteraction` objects index are obtained using the
        `truth_index_mode` attribute of the class.

        Parameters
        ----------
        obj : Union[FragmentBase, ParticleBase, InteractionBase]
            Fragment, Particle or Interaction object

        Results
        -------
        np.ndarray
           (N) Object index
        """
        if not obj.is_truth:
            return obj.index
        else:
            return getattr(obj, self.truth_index_mode)

    def get_points(self, obj):
        """Get a certain pre-defined point attribute of an object.

        The :class:`TruthFragment`, :class:`TruthParticle` and
        :class:`TruthInteraction` objects points are obtained using the
        `truth_point_mode` attribute of the class.

        Parameters
        ----------
        obj : Union[FragmentBase, ParticleBase, InteractionBase]
            Fragment, Particle or Interaction object

        Results
        -------
        np.ndarray
           (N, 3) Point coordinates
        """
        if not obj.is_truth:
            return obj.points
        else:
            return getattr(obj, self.truth_point_mode)

    @abstractmethod
    def process(self, data):
        """Place-holder method to be defined in each analysis script.

        Parameters
        ----------
        data : dict
            Filtered data dictionary for one entry
        """
        raise NotImplementedError("Must define the `process` function")



"""Diagnostic analaysis scripts.

This submodule is used to run basic diagnostics analyses, such as:
- Track energy reconstruction
- Track completeness
- Shower start dE/dx
- ...
"""

from .shower import *
from .track import *
